using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using NLog;
using System.IO;
using System.Collections;
using System.Data.SqlClient;

#region Class-Specific Usings


#endregion

namespace IdcgLibrary
{
    [Serializable]
    public class FloorItem
    {
        #region Constants

        public static string CLASS_NAME = typeof(FloorItem).ToString();

        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        #endregion

        #region Public Access

        #region Public Properties

        public string UpperIDValue
        {
            get { return mUpperID; }
            set { mUpperID = value; }
        }

        public string LowerIDValue
        {
            get { return mLowerID; }
            set { mLowerID = value; }
        }

        public OfficeList OfficesValue
        {
            get { return mOffices; }
            set { mOffices = value; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Set default values for all fields.
        /// </summary>
        public FloorItem()
        {
            Reset();
        }

        /// <summary>
        /// Set all fields according to arguments
        /// </summary>
        /// <param name=""></param>
        public FloorItem(string aUpperID, string aLowerID, OfficeList aOffices)
        {
            Set(aUpperID, aLowerID, aOffices);
        }

        /// <summary>
        /// Copy Constructor
        /// </summary>
        /// <param name=""></param>
        public FloorItem(FloorItem aItem)
        {
            Set(aItem);
        }

        #endregion

        #region Reset, Set

        /// <summary>
        /// Set fields to default values
        /// </summary>
        public void Reset()
        {
            mUpperID = "00000000-0000-0000-0000-000000000000";
            mLowerID = "00000000-0000-0000-0000-000000000000";
            mOffices = new OfficeList();
        }

        /// <summary>
        /// Set the values of every field according to arguments
        /// </summary>
        /// <param name=""></param>
        public void Set(string aUpperID, string aLowerID, OfficeList aOffices)
        {
            mUpperID = aUpperID;
            mLowerID = aLowerID;
            mOffices = new OfficeList(aOffices);
        }

        /// <summary>
        /// Copy our values from given FloorItem
        /// </summary>
        /// <param name="aFrom"></param>
        public void Set(FloorItem aFrom)
        {
            mUpperID = aFrom.mUpperID;
            mLowerID = aFrom.mLowerID;
            mOffices = new OfficeList(aFrom.mOffices);
        }

        #endregion

        #region Logging

        /// <summary>
        /// Return ToString for this class, or "null"
        /// </summary>
        /// <param name="aValue"></param>
        /// <returns></returns>
        public static string ToStringP(FloorItem aValue)
        {
            if (aValue == null)
                return "null";
            else
                return aValue.ToString();
        }

        /// <summary>
        /// Return fields of this object as a string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "mUpperID="+mUpperID+", mLowerID="+mLowerID+", mOffices= "+((mOffices == null) ? "null" : "(Cnt="+mOffices.GetCount()+")");
        }

        /// <summary>
        /// Log this object to nlog
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void Logit(LogLevel aLevel, string aIndent, string aHeader, IdLibrary.LogDocDepth aDepth)
        {
            string logString = LogitDetailToString(aLevel, aIndent, aHeader, aDepth);
            logger.Log(aLevel, logString);
        }


        /// <summary>
        /// Log the contents of this class to String at various levels of detail
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitDetailToString(LogLevel aLevel, string aIndent, string aHeader, IdLibrary.LogDocDepth aDepth)
        {
            #region Init

            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(logger, aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            #endregion

            // Minimum level does a one-line summary
            if (aDepth == IdLibrary.LogDocDepth.MINIMUM)
            {
                sb.Append(aIndent + aHeader + " " + ToString());

                return sb.ToString();
            }

            // Any level above MINIMUM
            sb.AppendLine(aIndent + aHeader + " FloorItem (" + aDepth + "):");

            sb.AppendLine(aIndent+"    "+"mUpperID="+mUpperID+", mLowerID="+mLowerID);

            // Child pages

            sb.AppendLine();
            if (mOffices == null || mOffices.GetCount() < 1)
                sb.Append(aIndent + "    " + "mOffices is null or empty");
            else
            {
                int childIdx;

                sb.AppendLine(aIndent + "    " + "mOffices (Count=" + mOffices.GetCount() + "):");

                for (childIdx = 0; childIdx < mOffices.GetCount(); childIdx++)
                {
                    string docValue;

                    switch (aDepth)
                    {
                        case IdLibrary.LogDocDepth.MAX:
                            docValue = mOffices[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:", aDepth);
                            break;

                        case IdLibrary.LogDocDepth.FULL:
                            docValue = mOffices[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:",
                                IdLibrary.LogDocDepth.MODERATE);
                            break;

                        case IdLibrary.LogDocDepth.MODERATE:
                            docValue = mOffices[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:",
                                IdLibrary.LogDocDepth.MINIMUM);
                            break;

                        default:
                            docValue = "unknown";   // should not happen
                            break;
                    }

                    sb.AppendLine(docValue);
                }
            }
            sb.AppendLine(aIndent + "    End of FloorItem");
            return sb.ToString();
        }


        #endregion

        #region I/O Routines

        /// <summary>
        /// Convert our class to an XML string via Serialize. Can throw exceptions
        /// </summary>
        /// <returns></returns>
        public string ToXML()
        {
            try
            {
                // Force UTF-8
                XmlSerializer serializer = new XmlSerializer(typeof(FloorItem));
                using (StringWriter writer = new IdLibrary.Utf8StringWriter())
                {
                    serializer.Serialize(writer, this);
                    return writer.ToString();
                }
            }
            catch (Exception ex)
            {
                logger.Error("ToXML: Exception '" + ex + "'");
                throw new Exception("ToXML", ex);
            }
        }

        /// <summary>
        /// Convert given xml string to our class. Throw exception on fail.
        /// </summary>
        /// <param name="xml"></param>
        /// <returns></returns>
        public FloorItem FromXML(string xml)
        {
            FloorItem ei = null;
            if ((xml != null) && (xml.Length > 0))
            {
                try
                {
                    XmlSerializer serializer = new XmlSerializer(typeof(FloorItem));
                    ei = (FloorItem)serializer.Deserialize(new StringReader(xml));
                }
                catch (Exception ex)
                {
                    logger.Error("FromXML: Got exception '" + ex + "'");
                    throw new Exception("FromXML", ex);
                }
            }
            return ei;
        }

        /// <summary>
        /// Write out our class to given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool WriteFile(string fileName)
        {
            try
            {
                StreamWriter writer = new StreamWriter(fileName);
                writer.Write(this.ToXML());
                writer.Close();
            }
            catch (Exception ex)
            {
                logger.Info("WriteFile: Got exception '" + ex + "'");
                return false;
            }
            // Success
            return true;
        }

        /// <summary>
        /// Read our class from given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool ReadFile(string fileName)
        {
            try
            {
                StreamReader reader = new StreamReader(fileName);
                string contents = reader.ReadToEnd();
                reader.Close();

                // Set all our values from what we got back
                FloorItem fi = FromXML(contents);
                this.Set(fi);
            }
            catch (Exception ex)
            {
                logger.Info("ReadFile: Got exception '" + ex + "'");
                return false;
            }

            // Success
            return true;
        }

        #endregion

        #region Tester

        /// <summary>
        /// Basic test routine for FloorItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool Tester()
        {
            string TEST_FILE = "testFloor.xml";
            bool result;
            FloorItem fi1, fi2, fi3;
            string upperID = Guid.NewGuid().ToString();

            logger.Info("");
            logger.Info("Tester:1:");

            // Test constructors
            try
            {
                fi1 = new FloorItem();
                fi2 = BuildSample(upperID, 1);
                fi3 = new FloorItem(fi2);
                fi3.UpperIDValue = "b8db4859-d558-46dd-b12a-8cb35cc15179";
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Constructor test got ex: " + ex);
                return false;
            }

            // Test logging
            try
            {
                logger.Info("Tester: ToStringP(fi2)=[" + ToStringP(fi2) + "]");
                logger.Info("");
                fi2.Logit(LogLevel.Info, "", "Tester: Logit without indent:", IdLibrary.LogDocDepth.MAX);
                fi3.Logit(LogLevel.Info, "   ", "Tester: Logit with indent:", IdLibrary.LogDocDepth.MINIMUM);
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Logit test got ex: " + ex);
                return false;
            }

            // Test Write and read
            try
            {
                logger.Info("");

                // Write out to XML
                if (fi2.WriteFile(TEST_FILE) == false)
                {
                    logger.Info("Tester: WriteFile(" + TEST_FILE + ") failed, returning false");
                    return false;
                }

                // Read in as different object
                result = fi1.ReadFile(TEST_FILE);

                if (result != true)
                {
                    logger.Info("Tester: Read " + TEST_FILE + " failed, returning false");
                    return false;
                }

                fi1.Logit(LogLevel.Info, "", "Tester: Read " + TEST_FILE + ":", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Read/Write test got ex: " + ex);
                return false;
            }

            // Success
            logger.Info("Tester: tests completed successfully");
            return true;
        }

        /// <summary>
        /// Return a sample record FloorItem.
        /// Return sample on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static FloorItem BuildSample(string aUpperValue, int aValue)
        {
            logger.Info("BuildSample: aUpperValue='"+aUpperValue+"'");
            string tempUpperID;
            string tempLowerID;
            OfficeList tempOffices;

            FloorItem rec;
            string lowerID = Guid.NewGuid().ToString();

            if (aValue == 1)
            {
                tempUpperID = "9edccf17-7c2d-4c6c-ac70-db61e6e8d071";
                tempLowerID = "32d8c36c-e0e1-40a5-8c5d-cfffd682f0a1";
                tempOffices = new OfficeList(OfficeList.BuildSample(lowerID, aValue));

                rec = new FloorItem(tempUpperID, tempLowerID, tempOffices);
                rec.UpperIDValue = aUpperValue;
                rec.LowerIDValue = lowerID;
            }
            else if (aValue == 2)
            {
                tempUpperID = "98b1ee08-b0c0-49fa-8d4d-d378e6d99051";
                tempLowerID = "35d8f26f-3d8f-46bf-9dd6-c8668ff74c42";
                tempOffices = new OfficeList(OfficeList.BuildSample(lowerID, aValue));

                rec = new FloorItem(tempUpperID, tempLowerID, tempOffices);
                rec.UpperIDValue = aUpperValue;
                rec.LowerIDValue = lowerID;
            }
            else if (aValue == 3)
            {
                tempUpperID = "b70d107a-1f14-4919-8b22-d0694765d109";
                tempLowerID = "8998da39-45a1-4291-b3d5-94a67751e9aa";
                tempOffices = new OfficeList(OfficeList.BuildSample(lowerID, aValue));

                rec = new FloorItem(tempUpperID, tempLowerID, tempOffices);
                rec.UpperIDValue = aUpperValue;
                rec.LowerIDValue = lowerID;
            }
            else
            {
                rec = new FloorItem("00000000-0000-0000-0000-000000000000", "00000000-0000-0000-0000-000000000000", new OfficeList());
            }

            return rec;
        }

        /// <summary>
        /// Drill down through sample record until non-class 2nd item and update it
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public bool UpdateSample()
        {
            return mOffices.UpdateSample();
        }

        #endregion

        #region Database Routines

        #region Create Table Routines

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public static IdLibrary.TriResult CreateTable(string aTableName)
        {
            return CreateTable(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("CreateTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return CreateTable(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("CreateTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("CreateTable: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return CreateTable(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Create given table in database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return CreateTable(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string cmdHeader = "CREATE TABLE "+aTableName+"(";
                string cmdArgs = "UpperID UNIQUEIDENTIFIER DEFAULT NEWID(), LowerID UNIQUEIDENTIFIER DEFAULT NEWID()";
                string cmdClose = ")";
                string command = cmdHeader + cmdArgs + cmdClose;
                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Trace("CreateTable: RunSql returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                    return tResult;

                // Add each class as a call for another table
                logger.Info("CreateTable:3: Calling OfficeList.CreateTable(OfficeList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = OfficeList.CreateTable("OfficeList", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("CreateTable: OfficeList.CreateTable() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("CreateTable: OfficeList.CreateTable() returned "+tResult.ToString());

                // Success (or NOTFOUND)
                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("CreateTable: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region Drop Table Routines

        /// <summary>
        /// Drop given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public static IdLibrary.TriResult DropTable(string aTableName)
        {
            return DropTable(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Drop given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult DropTable(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("CreateTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DropTable(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Drop given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult DropTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("DropTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("DropTable: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DropTable(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Drop given table in database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        public static IdLibrary.TriResult DropTable(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return DropTable(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Destroy given table in database.
        /// Return 1 on success, -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public static IdLibrary.TriResult DropTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string command = "DROP TABLE dbo."+aTableName;
                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Trace("DropTable: RunSql returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                    return tResult;

                // Add each class as a call for another table
                logger.Info("DropTable:3: Calling OfficeList.DropTable(OfficeList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = OfficeList.DropTable("OfficeList", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DropTable: OfficeList.DropTable() failed("+tResult.ToString()+")");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DropTable: OfficeList.DropTable() returned "+tResult.ToString());

                // Success (SUC or NOTFOUND)
                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("DropTable: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Insert Routines

        /// <summary>
        /// Insert this record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        public IdLibrary.TriResult InsertRecord(string aTableName)
        {
            return InsertRecord(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Insert this record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("InsertRecord: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return InsertRecord(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Insert this record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        /// <param name="aConnection">Connection to use</param>
        /// <param name="aTrans">Transaction to use</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("InsertRecord: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("InsertRecord: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return InsertRecord(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Insert this record into database using supplied connection credentials.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return InsertRecord(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Insert this record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string cmdHeader = "INSERT INTO "+aTableName+" (";
                string columnArgs = "UpperID, LowerID";
                string colClose = ") Values (";
                string valueArgs = "'"+mUpperID+"', '"+mLowerID+"'";
                string cmdClose = ");";
                string command = cmdHeader + columnArgs + colClose + valueArgs + cmdClose;

                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Info("InsertRecord:2: RunSqlNonQuery returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("InsertRecord: SQL INSERT command failed");
                    return IdLibrary.TriResult.FAIL;
                }

                // Add each class as a call for another table
                logger.Info("InsertRecord:3: Calling mOffices.InsertRecord(OfficeList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = mOffices.InsertRecord("OfficeList", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("InsertRecord: "+mOffices+".InsertRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("InsertRecord: "+mOffices+".InsertRecord() returned "+tResult.ToString());

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("InsertRecord: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Update Routines

        /// <summary>
        /// Update all fields of this record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName)
        {
            return UpdateRecordByLowerId(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Update all fields of this record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("UpdateRecordByLowerId: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return UpdateRecordByLowerId(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Update all fields of this record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        /// <param name="aTrans">Transaction to use</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("UpdateRecordByLowerId: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("UpdateRecordByLowerId: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return UpdateRecordByLowerId(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Update all fields of this record in database, using provided connection info.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aDbName">Database Name to use</param>
        /// <param name="aUserID">UserID to log into DB</param>
        /// <param name="aPassword">Password to log into DB</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return UpdateRecordByLowerId(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Update all fields of this record in database. Should only be called by ICDG routines
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string cmdHeader = "UPDATE "+aTableName+" SET ";
                string columnArgs = "UpperID = '"+mUpperID+"', LowerID = '"+mLowerID+"'";
                string colClose = " WHERE LowerID =";
                string valueArgs = "'"+mLowerID+"'";
                string cmdClose = ";";
                string command = cmdHeader + columnArgs + colClose + valueArgs + cmdClose;

                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Trace("UpdateRecordByLowerId: RunSqlNonQuery returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("UpdateRecordByLowerId: SQL UPDATE command failed");
                    return IdLibrary.TriResult.FAIL;
                }

                // Add each class as a call for another table
                logger.Info("UpdateRecordByLowerId:3: Calling mOffices.UpdateRecordByLowerId(OfficeList, aConnection, aTrans)");
                tResult = mOffices.UpdateRecordByLowerId("OfficeList", aConnection, aTrans);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("UpdateRecordByLowerId: "+mOffices+".UpdateRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("UpdateRecordByLowerId: "+mOffices+".UpdateRecord() returned "+tResult.ToString());

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("UpdateRecordByLowerId: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Delete Routines

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, string aWhereClause)
        {
            return DeleteRecordByWhere(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, aWhereClause);
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            if (aConnection == null)
            {
                logger.Error("DeleteRecordByWhere: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DeleteRecordByWhere(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, aWhereClause);
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aWhereClause)
        {
            if (aConnection == null)
            {
                logger.Error("DeleteRecordByWhere: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("DeleteRecordByWhere: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DeleteRecordByWhere(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, aWhereClause);
        }

        /// <summary>
        /// Delete this record from database using supplied connection credentials.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, string aDbName, string aUserID, string aPassword, string aWhereClause)
        {
            return DeleteRecordByWhere(aTableName, null, null, aDbName, aUserID, aPassword, aWhereClause);
        }

        /// <summary>
        /// Delete this record from database. Should only be called by IDCG.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword, string aWhereClause)
        {
            IdLibrary.TriResult tResult;

            try
            {
                // Add each class as a call for another table
                logger.Info("DeleteRecordByWhere:3: Calling mOffices.DeleteRecordByLowerId(OfficeList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = mOffices.DeleteRecordByLowerId("OfficeList", aConnection, aTrans, aDbName, aUserID, aPassword);

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DeleteRecordByWhere: "+mOffices+".DeleteRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DeleteRecordByWhere: "+mOffices+".DeleteRecord() returned "+tResult);

                string cmdHeader = "DELETE FROM "+aTableName+" WHERE ";
                string valueArgs = aWhereClause;
                string cmdClose = ";";
                string command = cmdHeader + valueArgs + cmdClose;

                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DeleteRecordByWhere: RunSqlNonQuery failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DeleteRecordByWhere: RunSql returned "+tResult.ToString());
                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("DeleteRecordByWhere: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName)
        {
            return DeleteRecordByWhere(aTableName, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, SqlConnection aConnection)
        {
            return DeleteRecordByWhere(aTableName, aConnection, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use</param>
        /// <param name="aTrans">Transaction to use</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            return DeleteRecordByWhere(aTableName, aConnection, aTrans, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aDbName">Database Name to connect to</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return DeleteRecordByWhere(aTableName, aDbName, aUserID, aPassword, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete this record from database. Should only be called by IDCG
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            return DeleteRecordByWhere(aTableName, aConnection, aTrans, aDbName, aUserID, aPassword, "LowerID = '"+ mLowerID + "'");
        }

        #endregion

        #region DB Record Select Open/Read Routines

        /// <summary>
        /// Create and open the SQL select to get fields of this record.
        /// Return reader on success, null on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public SqlDataReader SqlOpen(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            SqlDataReader myReader = null;

            try
            {
                string cmdHeader = "SELECT ";
                string columnArgs = "UpperID, LowerID";
                string colClose = " FROM "+aTableName+" WHERE " + aWhereClause + ";";
                string command = cmdHeader + columnArgs + colClose;

                logger.Info("SqlOpen: command=["+command+"]");
                SqlCommand myCommand = new SqlCommand(command, aConnection);
                myReader = myCommand.ExecuteReader();
                return myReader;
            }
            catch (Exception ex)
            {
                logger.Error("SqlOpen: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return null;
            }
        }

        /// <summary>
        /// Read in the next sql record into this Item.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aReader">Reader to read from</param>
        public IdLibrary.TriResult SqlRead(SqlDataReader myReader)
        {
            IdLibrary.TriResult tResult = IdLibrary.TriResult.SUCCESS;
            bool result;

            try
            {
                result = myReader.Read();
                if (result == false)
                    return IdLibrary.TriResult.NOTFOUND;

                mUpperID = myReader.GetGuid(0).ToString();
                mLowerID = myReader.GetGuid(1).ToString();
                mOffices = new OfficeList();

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("SqlRead: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Select Routines

        /// <summary>
        /// Select 1 record from database according to LowerID of this record.
        /// Return SUCCESS, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult SqlSelectByLowerId(string aTableName, SqlConnection aConnection)
        {
            return SqlSelectByWhere(aTableName, aConnection, "LowerID = '"+mLowerID+"'");
        }

        /// <summary>
        /// Select 1 record from database according to given where clause.
        /// Return SUCCESS, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult SqlSelectByWhere(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            SqlDataReader myReader = null;
            IdLibrary.TriResult tResult;

            try
            {
                // Create and open the appropriate select
                myReader = SqlOpen(aTableName, aConnection, aWhereClause);
                if (myReader == null)
                    return IdLibrary.TriResult.FAIL;

                // Since this is an item, get exactly 1 record
                tResult = SqlRead(myReader);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                if (tResult != IdLibrary.TriResult.SUCCESS)
                    return tResult;

                return SqlSelectSubrecords(aConnection);
            }
            catch (Exception ex)
            {
                logger.Error("SqlSelectByWhere: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return IdLibrary.TriResult.FAIL;
            }
        }

        /// <summary>
        /// Load subclasses with their sql data.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        public IdLibrary.TriResult SqlSelectSubrecords(SqlConnection aConnection)
        {
            try
            {
                IdLibrary.TriResult tResult = IdLibrary.TriResult.SUCCESS;

                // Add each class as a call for another table
                string whereClause = "UpperID = '"+mLowerID+"'";
                logger.Info("SqlSelectSubrecords:3: Calling mOffices.SqlSelectByWhere(OfficeList, aConnection, whereClause)");
                tResult = mOffices.SqlSelectByWhere("OfficeList", aConnection, whereClause);

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("SqlSelectSubrecords: mOffices.SqlSelectByWhere() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("SqlSelectSubrecords: mOffices.SqlSelectByWhere() returned "+tResult);

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("SqlSelectSubrecords: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #endregion

        #region TestDatabase

        /// <summary>
        /// Basic database test routine for FloorItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabase(bool aDeleteAtEnd, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult, readResult;
            SqlConnection myConnection = null;
            SqlTransaction myTrans = null;
            FloorItem fi1, readItem;
            string whereClause;
            string upperID = Guid.NewGuid().ToString();

            logger.Info("");
            logger.Info("TestDatabase:1:");

            // Test building sample record
            try
            {
                fi1 = BuildSample(upperID, 1);
                fi1.Logit(LogLevel.Info, "", "TestDatabase: Sample record:", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: Build sample record test got ex: " + ex);
                return false;
            }

            // Test database open
            logger.Info("TestDatabase:2: Testing OpenDb");
            try
            {
                myConnection = IdLibrary.OpenDbConnection(null, aDbName, aUserID, aPassword);
                if (myConnection == null)
                {
                    logger.Error("TestDatabase: Can't open connection to database '"+aDbName+"'");
                    return false;
                }
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: Open test got ex: " + ex);
                return false;
            }

            // Test drop existing table, ignore table not present
            logger.Info("TestDatabase:3: Testing initial DropTable");
            try
            {
                // Begin transaction
                myTrans = myConnection.BeginTransaction();

                tResult = FloorItem.DropTable("FloorItem", myConnection, myTrans);
                if (IdLibrary.AnyIdError(tResult))
                {
                    logger.Info("TestDatabase: Initial drop of 'FloorItem' returned "+tResult+", ignore");
                }
                else
                {
                    logger.Info("TestDatabase: Initial drop of 'FloorItem' returned "+tResult);
                }

                // Commit on success
                myTrans.Commit(); myTrans = null;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: DropTable test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            // Test create new table, will fail if table cannot be created
            logger.Info("TestDatabase:4: Testing CreateTable");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = FloorItem.CreateTable("FloorItem", myConnection, myTrans);
                logger.Info("TestDatabase: CreateTable test returned " + tResult.ToString());

                if (IdLibrary.AnyIdError(tResult))
                {
                    logger.Info("TestDatabase: Create of 'FloorItem' failed (Cnt="+tResult.ToString()+")");
                    myTrans.Rollback(); myTrans = null;
                    myConnection.Close();
                    return false;
                }
                else
                {
                    logger.Info("TestDatabase: Create of 'FloorItem' success");
                }

                // Commit on success
                myTrans.Commit(); myTrans = null;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: CreateTable test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            // Test inserting record into table, will fail if insert fails
            logger.Info("TestDatabase:5: Testing InsertRecord");
            try
            {
                readItem = new FloorItem();
                readItem.mLowerID = fi1.mLowerID;
                readResult = readItem.SqlSelectByLowerId("FloorItem", myConnection);
                logger.Info("TestDatabase:5.1: SqlSelectByLowerId returned " + readResult.ToString());

                if (readResult != IdLibrary.TriResult.NOTFOUND)
                {
                    logger.Info("TestDatabase:5.2: Already a "+fi1.mLowerID+" record in db. Insert will fail");
                    myConnection.Close();
                    return false;
                }

                myTrans = myConnection.BeginTransaction();
                tResult = fi1.InsertRecord("FloorItem", myConnection, myTrans);
                if (IdLibrary.AnyIdError(tResult))
                {
                    logger.Info("TestDatabase: Insert record 1 for 'FloorItem' failed (Cnt="+tResult.ToString()+")");
                    myTrans.Rollback(); myTrans = null;
                    myConnection.Close();
                    return false;
                }

                myTrans.Commit(); myTrans = null;
                readItem = new FloorItem();
                readItem.mLowerID = fi1.mLowerID;
                readResult = readItem.SqlSelectByLowerId("FloorItem", myConnection);

                if (readResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabase:5.4: Failed to get "+fi1.mLowerID+" record from db.");
                    myConnection.Close();
                    return false;
                }

                readItem.Logit(LogLevel.Info, "", "TestDatabase: After insert:", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: InsertRecord test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            // Test updating record into table, will fail if insert fails
            logger.Info("TestDatabase:6: Testing UpdateRecordbyLowerId");
            try
            {
                whereClause = "LowerID = '"+fi1.mLowerID+"'";

                if (fi1.UpdateSample() == false)
                {
                    logger.Info("TestDatabase: Update Sample for 'FloorItem' failed ");
                    myConnection.Close();
                    return false;
                }

                fi1.Logit(LogLevel.Info, "", "TestDatabase: After update sample:", IdLibrary.LogDocDepth.MAX);
                myTrans = myConnection.BeginTransaction();

                tResult = fi1.UpdateRecordByLowerId("FloorItem", myConnection, myTrans);
                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabase: Update record 1 for 'FloorItem' failed (Cnt="+tResult.ToString()+")");
                    myTrans.Rollback();
                    myConnection.Close();
                    return false;
                }

                myTrans.Commit(); myTrans = null;
                readItem = new FloorItem();
                readItem.mLowerID = fi1.mLowerID;
                readResult = readItem.SqlSelectByWhere("FloorItem", myConnection, whereClause);

                if (readResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabase:6.4: Failed to get "+fi1.mLowerID+" record from db.");
                    myConnection.Close();
                    return false;
                }

                readItem.Logit(LogLevel.Info, "", "TestDatabase: After update:", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: UpdateRecord test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            if (aDeleteAtEnd == true)
            {
                if (TestDatabaseFinalRecordDelete(myConnection, fi1, whereClause) == false)
                {
                    logger.Info("TestDatabase: TestDatabaseFinalRecordDelete() failed");
                    myConnection.Close();
                    return false;
                }

                if (TestDatabaseDropTable(myConnection) == false)
                {
                    logger.Info("TestDatabase: TestDatabaseDropTable() failed");
                    myConnection.Close();
                    return false;
                }
            }

            // Success
            logger.Info("TestDatabase: tests completed successfully");
            myConnection.Close();
            return true;
        }

        /// <summary>
        /// database final table delete test routine for FloorItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabaseFinalRecordDelete(SqlConnection myConnection, FloorItem fi1, string whereClause)
        {
            IdLibrary.TriResult tResult;
            FloorItem readItem;
            SqlTransaction myTrans = null;

            // Test deleting record into table, will fail if insert fails
            logger.Info("TestDatabaseFinalRecordDelete:1: Testing DeleteRecordByWhere");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = fi1.DeleteRecordByWhere("FloorItem", myConnection, myTrans, whereClause);
                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabaseFinalTableDelete: Delete record for 'FloorItem' failed (Cnt="+tResult.ToString()+")");
                    if (myTrans != null)
                    {
                        myTrans.Rollback(); myTrans = null;
                    }
                    return false;
                }

                myTrans.Commit(); myTrans = null;
                readItem = new FloorItem();
                readItem.mLowerID = fi1.mLowerID;
                tResult = readItem.SqlSelectByWhere("FloorItem", myConnection, whereClause);

                if (tResult != IdLibrary.TriResult.NOTFOUND)
                {
                    readItem.Logit(LogLevel.Error, "", "TestDatabaseFinalRecordDelete: Record was supposed to be deleted:", IdLibrary.LogDocDepth.MAX);
                    return false;
                }

                return true;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabaseFinalRecordDelete: DeleteRecordByWhere test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                return false;
            }
        }

        /// <summary>
        /// database drop test routine for FloorItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabaseDropTable(SqlConnection myConnection)
        {
            IdLibrary.TriResult tResult;
            SqlTransaction myTrans = null;

            // Test drop existing table, will fail if table not present
            logger.Info("TestDatabaseDropTable:1: Testing final DropTable");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = FloorItem.DropTable("FloorItem", myConnection, myTrans);
                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabaseDropTable: Final drop of 'FloorItem' returned "+tResult.ToString()+", returning false");
                    if (myTrans != null)
                    {
                        myTrans.Rollback(); myTrans = null;
                    }
                    return false;
                }

                // Success
                logger.Info("TestDatabaseDropTable: test completed successfully");
                myTrans.Commit(); myTrans = null;
                return true;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabaseDropTable: got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                return false;
            }
        }

        #endregion

        #region Class-Specific


        #endregion


        #endregion

        #region Private Access

        #region Private Members

        private string mUpperID;
        private string mLowerID;
        private OfficeList mOffices;

        #endregion

        #endregion
    }
}
