using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using NLog;
using System.IO;
using System.Collections;
using System.Data.SqlClient;

#region Class-Specific Usings


#endregion

namespace IdcgLibrary
{
    [Serializable]
    public class OfficeItem
    {
        #region Constants

        public static string CLASS_NAME = typeof(OfficeItem).ToString();

        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        #endregion

        #region Public Access

        #region Public Properties

        public string UpperIDValue
        {
            get { return mUpperID; }
            set { mUpperID = value; }
        }

        public string LowerIDValue
        {
            get { return mLowerID; }
            set { mLowerID = value; }
        }

        public ComputerList ComputersValue
        {
            get { return mComputers; }
            set { mComputers = value; }
        }

        public PhoneList PhonesValue
        {
            get { return mPhones; }
            set { mPhones = value; }
        }

        public PersonItem OccupantValue
        {
            get { return mOccupant; }
            set { mOccupant = value; }
        }

        public string RoomNumberValue
        {
            get { return mRoomNumber; }
            set { mRoomNumber = value; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Set default values for all fields.
        /// </summary>
        public OfficeItem()
        {
            Reset();
        }

        /// <summary>
        /// Set all fields according to arguments
        /// </summary>
        /// <param name=""></param>
        public OfficeItem(string aUpperID, string aLowerID, ComputerList aComputers, PhoneList aPhones, PersonItem aOccupant, string aRoomNumber)
        {
            Set(aUpperID, aLowerID, aComputers, aPhones, aOccupant, aRoomNumber);
        }

        /// <summary>
        /// Copy Constructor
        /// </summary>
        /// <param name=""></param>
        public OfficeItem(OfficeItem aItem)
        {
            Set(aItem);
        }

        #endregion

        #region Reset, Set

        /// <summary>
        /// Set fields to default values
        /// </summary>
        public void Reset()
        {
            mUpperID = "00000000-0000-0000-0000-000000000000";
            mLowerID = "00000000-0000-0000-0000-000000000000";
            mComputers = new ComputerList();
            mPhones = new PhoneList();
            mOccupant = new PersonItem();
            mRoomNumber = "unused";
        }

        /// <summary>
        /// Set the values of every field according to arguments
        /// </summary>
        /// <param name=""></param>
        public void Set(string aUpperID, string aLowerID, ComputerList aComputers, PhoneList aPhones, PersonItem aOccupant, string aRoomNumber)
        {
            mUpperID = aUpperID;
            mLowerID = aLowerID;
            mComputers = new ComputerList(aComputers);
            mPhones = new PhoneList(aPhones);
            mOccupant = new PersonItem(aOccupant);
            mRoomNumber = aRoomNumber;
        }

        /// <summary>
        /// Copy our values from given OfficeItem
        /// </summary>
        /// <param name="aFrom"></param>
        public void Set(OfficeItem aFrom)
        {
            mUpperID = aFrom.mUpperID;
            mLowerID = aFrom.mLowerID;
            mComputers = new ComputerList(aFrom.mComputers);
            mPhones = new PhoneList(aFrom.mPhones);
            mOccupant = new PersonItem(aFrom.mOccupant);
            mRoomNumber = aFrom.mRoomNumber;
        }

        #endregion

        #region Logging

        /// <summary>
        /// Return ToString for this class, or "null"
        /// </summary>
        /// <param name="aValue"></param>
        /// <returns></returns>
        public static string ToStringP(OfficeItem aValue)
        {
            if (aValue == null)
                return "null";
            else
                return aValue.ToString();
        }

        /// <summary>
        /// Return fields of this object as a string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "mUpperID="+mUpperID+", mLowerID="+mLowerID+", mComputers= "+((mComputers == null) ? "null" : "(Cnt="+mComputers.GetCount()+")")+", mPhones= "+((mPhones == null) ? "null" : "(Cnt="+mPhones.GetCount()+")")+", mOccupant= "+((mOccupant == null) ? "null" : "set")+", mRoomNumber='"+mRoomNumber+"'";
        }

        /// <summary>
        /// Log this object to nlog
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void Logit(LogLevel aLevel, string aIndent, string aHeader, IdLibrary.LogDocDepth aDepth)
        {
            string logString = LogitDetailToString(aLevel, aIndent, aHeader, aDepth);
            logger.Log(aLevel, logString);
        }


        /// <summary>
        /// Log the contents of this class to String at various levels of detail
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitDetailToString(LogLevel aLevel, string aIndent, string aHeader, IdLibrary.LogDocDepth aDepth)
        {
            #region Init

            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(logger, aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            #endregion

            // Minimum level does a one-line summary
            if (aDepth == IdLibrary.LogDocDepth.MINIMUM)
            {
                sb.Append(aIndent + aHeader + " " + ToString());

                return sb.ToString();
            }

            // Any level above MINIMUM
            sb.AppendLine(aIndent + aHeader + " OfficeItem (" + aDepth + "):");

            sb.AppendLine(aIndent+"    "+"mUpperID="+mUpperID+", mLowerID="+mLowerID);
            sb.AppendLine(aIndent + "    "+"mRoomNumber='"+mRoomNumber+"'");

            // Child pages

            sb.AppendLine();
            if (mComputers == null || mComputers.GetCount() < 1)
                sb.Append(aIndent + "    " + "mComputers is null or empty");
            else
            {
                int childIdx;

                sb.AppendLine(aIndent + "    " + "mComputers (Count=" + mComputers.GetCount() + "):");

                for (childIdx = 0; childIdx < mComputers.GetCount(); childIdx++)
                {
                    string docValue;

                    switch (aDepth)
                    {
                        case IdLibrary.LogDocDepth.MAX:
                            docValue = mComputers[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:", aDepth);
                            break;

                        case IdLibrary.LogDocDepth.FULL:
                            docValue = mComputers[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:",
                                IdLibrary.LogDocDepth.MODERATE);
                            break;

                        case IdLibrary.LogDocDepth.MODERATE:
                            docValue = mComputers[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:",
                                IdLibrary.LogDocDepth.MINIMUM);
                            break;

                        default:
                            docValue = "unknown";   // should not happen
                            break;
                    }

                    sb.AppendLine(docValue);
                }
            }

            sb.AppendLine();
            if (mPhones == null || mPhones.GetCount() < 1)
                sb.Append(aIndent + "    " + "mPhones is null or empty");
            else
            {
                int childIdx;

                sb.AppendLine(aIndent + "    " + "mPhones (Count=" + mPhones.GetCount() + "):");

                for (childIdx = 0; childIdx < mPhones.GetCount(); childIdx++)
                {
                    string docValue;

                    switch (aDepth)
                    {
                        case IdLibrary.LogDocDepth.MAX:
                            docValue = mPhones[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:", aDepth);
                            break;

                        case IdLibrary.LogDocDepth.FULL:
                            docValue = mPhones[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:",
                                IdLibrary.LogDocDepth.MODERATE);
                            break;

                        case IdLibrary.LogDocDepth.MODERATE:
                            docValue = mPhones[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:",
                                IdLibrary.LogDocDepth.MINIMUM);
                            break;

                        default:
                            docValue = "unknown";   // should not happen
                            break;
                    }

                    sb.AppendLine(docValue);
                }
            }

            sb.AppendLine();
            if (mOccupant == null )
                sb.Append(aIndent + "    " + "mOccupant is null or empty");
            else
            {
                sb.AppendLine(aIndent + "    " + "mOccupant:");

                    string docValue;

                    switch (aDepth)
                    {
                        case IdLibrary.LogDocDepth.MAX:
                            docValue = mOccupant.LogitDetailToString(aLevel, aIndent + "    ", "", aDepth);
                            break;

                        case IdLibrary.LogDocDepth.FULL:
                            docValue = mOccupant.LogitDetailToString(aLevel, aIndent + "    ", "",
                                IdLibrary.LogDocDepth.MODERATE);
                            break;

                        case IdLibrary.LogDocDepth.MODERATE:
                            docValue = mOccupant.LogitDetailToString(aLevel, aIndent + "    ", "",
                                IdLibrary.LogDocDepth.MINIMUM);
                            break;

                        default:
                            docValue = "unknown";   // should not happen
                            break;
                    }

                    sb.AppendLine(docValue);
            }
            sb.AppendLine(aIndent + "    End of OfficeItem");
            return sb.ToString();
        }


        #endregion

        #region I/O Routines

        /// <summary>
        /// Convert our class to an XML string via Serialize. Can throw exceptions
        /// </summary>
        /// <returns></returns>
        public string ToXML()
        {
            try
            {
                // Force UTF-8
                XmlSerializer serializer = new XmlSerializer(typeof(OfficeItem));
                using (StringWriter writer = new IdLibrary.Utf8StringWriter())
                {
                    serializer.Serialize(writer, this);
                    return writer.ToString();
                }
            }
            catch (Exception ex)
            {
                logger.Error("ToXML: Exception '" + ex + "'");
                throw new Exception("ToXML", ex);
            }
        }

        /// <summary>
        /// Convert given xml string to our class. Throw exception on fail.
        /// </summary>
        /// <param name="xml"></param>
        /// <returns></returns>
        public OfficeItem FromXML(string xml)
        {
            OfficeItem ei = null;
            if ((xml != null) && (xml.Length > 0))
            {
                try
                {
                    XmlSerializer serializer = new XmlSerializer(typeof(OfficeItem));
                    ei = (OfficeItem)serializer.Deserialize(new StringReader(xml));
                }
                catch (Exception ex)
                {
                    logger.Error("FromXML: Got exception '" + ex + "'");
                    throw new Exception("FromXML", ex);
                }
            }
            return ei;
        }

        /// <summary>
        /// Write out our class to given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool WriteFile(string fileName)
        {
            try
            {
                StreamWriter writer = new StreamWriter(fileName);
                writer.Write(this.ToXML());
                writer.Close();
            }
            catch (Exception ex)
            {
                logger.Info("WriteFile: Got exception '" + ex + "'");
                return false;
            }
            // Success
            return true;
        }

        /// <summary>
        /// Read our class from given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool ReadFile(string fileName)
        {
            try
            {
                StreamReader reader = new StreamReader(fileName);
                string contents = reader.ReadToEnd();
                reader.Close();

                // Set all our values from what we got back
                OfficeItem fi = FromXML(contents);
                this.Set(fi);
            }
            catch (Exception ex)
            {
                logger.Info("ReadFile: Got exception '" + ex + "'");
                return false;
            }

            // Success
            return true;
        }

        #endregion

        #region Tester

        /// <summary>
        /// Basic test routine for OfficeItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool Tester()
        {
            string TEST_FILE = "testOffice.xml";
            bool result;
            OfficeItem fi1, fi2, fi3;
            string upperID = Guid.NewGuid().ToString();

            logger.Info("");
            logger.Info("Tester:1:");

            // Test constructors
            try
            {
                fi1 = new OfficeItem();
                fi2 = BuildSample(upperID, 1);
                fi3 = new OfficeItem(fi2);
                fi3.UpperIDValue = "68739513-84e7-4e82-a26b-0c12015c4b2c";
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Constructor test got ex: " + ex);
                return false;
            }

            // Test logging
            try
            {
                logger.Info("Tester: ToStringP(fi2)=[" + ToStringP(fi2) + "]");
                logger.Info("");
                fi2.Logit(LogLevel.Info, "", "Tester: Logit without indent:", IdLibrary.LogDocDepth.MAX);
                fi3.Logit(LogLevel.Info, "   ", "Tester: Logit with indent:", IdLibrary.LogDocDepth.MINIMUM);
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Logit test got ex: " + ex);
                return false;
            }

            // Test Write and read
            try
            {
                logger.Info("");

                // Write out to XML
                if (fi2.WriteFile(TEST_FILE) == false)
                {
                    logger.Info("Tester: WriteFile(" + TEST_FILE + ") failed, returning false");
                    return false;
                }

                // Read in as different object
                result = fi1.ReadFile(TEST_FILE);

                if (result != true)
                {
                    logger.Info("Tester: Read " + TEST_FILE + " failed, returning false");
                    return false;
                }

                fi1.Logit(LogLevel.Info, "", "Tester: Read " + TEST_FILE + ":", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Read/Write test got ex: " + ex);
                return false;
            }

            // Success
            logger.Info("Tester: tests completed successfully");
            return true;
        }

        /// <summary>
        /// Return a sample record OfficeItem.
        /// Return sample on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static OfficeItem BuildSample(string aUpperValue, int aValue)
        {
            logger.Info("BuildSample: aUpperValue='"+aUpperValue+"'");
            string tempUpperID;
            string tempLowerID;
            ComputerList tempComputers;
            PhoneList tempPhones;
            PersonItem tempOccupant;
            string tempRoomNumber;

            OfficeItem rec;
            string lowerID = Guid.NewGuid().ToString();

            if (aValue == 1)
            {
                tempUpperID = "da6d64ca-4fa1-4b88-8b63-f5120c0c1b3f";
                tempLowerID = "2d19e8ae-74c5-43b9-b789-d67dc0353fe1";
                tempComputers = new ComputerList(ComputerList.BuildSample(lowerID, aValue));
                tempPhones = new PhoneList(PhoneList.BuildSample(lowerID, aValue));
                tempOccupant = new PersonItem(PersonItem.BuildSample(lowerID, aValue));
                tempRoomNumber = "aaa";

                rec = new OfficeItem(tempUpperID, tempLowerID, tempComputers, tempPhones, tempOccupant, tempRoomNumber);
                rec.UpperIDValue = aUpperValue;
                rec.LowerIDValue = lowerID;
            }
            else if (aValue == 2)
            {
                tempUpperID = "288c8fb7-5842-4e33-bb1c-2d6f3641aaab";
                tempLowerID = "71b25bbe-f6d8-4195-9f37-c27750232d12";
                tempComputers = new ComputerList(ComputerList.BuildSample(lowerID, aValue));
                tempPhones = new PhoneList(PhoneList.BuildSample(lowerID, aValue));
                tempOccupant = new PersonItem(PersonItem.BuildSample(lowerID, aValue));
                tempRoomNumber = "bbb";

                rec = new OfficeItem(tempUpperID, tempLowerID, tempComputers, tempPhones, tempOccupant, tempRoomNumber);
                rec.UpperIDValue = aUpperValue;
                rec.LowerIDValue = lowerID;
            }
            else if (aValue == 3)
            {
                tempUpperID = "f68d042f-b17d-468d-8849-7f629776764f";
                tempLowerID = "1dcc89ad-093a-4e1b-b4f4-91c428e29a14";
                tempComputers = new ComputerList(ComputerList.BuildSample(lowerID, aValue));
                tempPhones = new PhoneList(PhoneList.BuildSample(lowerID, aValue));
                tempOccupant = new PersonItem(PersonItem.BuildSample(lowerID, aValue));
                tempRoomNumber = "ccc";

                rec = new OfficeItem(tempUpperID, tempLowerID, tempComputers, tempPhones, tempOccupant, tempRoomNumber);
                rec.UpperIDValue = aUpperValue;
                rec.LowerIDValue = lowerID;
            }
            else
            {
                rec = new OfficeItem("00000000-0000-0000-0000-000000000000", "00000000-0000-0000-0000-000000000000", new ComputerList(), new PhoneList(), new PersonItem(), "unused");
            }

            return rec;
        }

        /// <summary>
        /// Drill down through sample record until non-class 2nd item and update it
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public bool UpdateSample()
        {
            return mComputers.UpdateSample();
        }

        #endregion

        #region Database Routines

        #region Create Table Routines

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public static IdLibrary.TriResult CreateTable(string aTableName)
        {
            return CreateTable(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("CreateTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return CreateTable(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("CreateTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("CreateTable: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return CreateTable(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Create given table in database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return CreateTable(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string cmdHeader = "CREATE TABLE "+aTableName+"(";
                string cmdArgs = "UpperID UNIQUEIDENTIFIER DEFAULT NEWID(), LowerID UNIQUEIDENTIFIER DEFAULT NEWID(), RoomNumber varchar(255)";
                string cmdClose = ")";
                string command = cmdHeader + cmdArgs + cmdClose;
                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Trace("CreateTable: RunSql returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                    return tResult;

                // Add each class as a call for another table
                logger.Info("CreateTable:3: Calling ComputerList.CreateTable(ComputerList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = ComputerList.CreateTable("ComputerList", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("CreateTable: ComputerList.CreateTable() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("CreateTable: ComputerList.CreateTable() returned "+tResult.ToString());

                logger.Info("CreateTable:3: Calling PhoneList.CreateTable(PhoneList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = PhoneList.CreateTable("PhoneList", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("CreateTable: PhoneList.CreateTable() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("CreateTable: PhoneList.CreateTable() returned "+tResult.ToString());

                logger.Info("CreateTable:3: Calling PersonItem.CreateTable(PersonItem, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = PersonItem.CreateTable("PersonItem", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("CreateTable: PersonItem.CreateTable() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("CreateTable: PersonItem.CreateTable() returned "+tResult.ToString());

                // Success (or NOTFOUND)
                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("CreateTable: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region Drop Table Routines

        /// <summary>
        /// Drop given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public static IdLibrary.TriResult DropTable(string aTableName)
        {
            return DropTable(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Drop given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult DropTable(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("CreateTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DropTable(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Drop given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult DropTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("DropTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("DropTable: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DropTable(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Drop given table in database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        public static IdLibrary.TriResult DropTable(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return DropTable(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Destroy given table in database.
        /// Return 1 on success, -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public static IdLibrary.TriResult DropTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string command = "DROP TABLE dbo."+aTableName;
                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Trace("DropTable: RunSql returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                    return tResult;

                // Add each class as a call for another table
                logger.Info("DropTable:3: Calling ComputerList.DropTable(ComputerList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = ComputerList.DropTable("ComputerList", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DropTable: ComputerList.DropTable() failed("+tResult.ToString()+")");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DropTable: ComputerList.DropTable() returned "+tResult.ToString());

                logger.Info("DropTable:3: Calling PhoneList.DropTable(PhoneList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = PhoneList.DropTable("PhoneList", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DropTable: PhoneList.DropTable() failed("+tResult.ToString()+")");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DropTable: PhoneList.DropTable() returned "+tResult.ToString());

                logger.Info("DropTable:3: Calling PersonItem.DropTable(PersonItem, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = PersonItem.DropTable("PersonItem", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DropTable: PersonItem.DropTable() failed("+tResult.ToString()+")");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DropTable: PersonItem.DropTable() returned "+tResult.ToString());

                // Success (SUC or NOTFOUND)
                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("DropTable: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Insert Routines

        /// <summary>
        /// Insert this record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        public IdLibrary.TriResult InsertRecord(string aTableName)
        {
            return InsertRecord(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Insert this record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("InsertRecord: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return InsertRecord(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Insert this record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        /// <param name="aConnection">Connection to use</param>
        /// <param name="aTrans">Transaction to use</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("InsertRecord: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("InsertRecord: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return InsertRecord(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Insert this record into database using supplied connection credentials.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return InsertRecord(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Insert this record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string cmdHeader = "INSERT INTO "+aTableName+" (";
                string columnArgs = "UpperID, LowerID, RoomNumber";
                string colClose = ") Values (";
                string valueArgs = "'"+mUpperID+"', '"+mLowerID+"', '"+mRoomNumber+"'";
                string cmdClose = ");";
                string command = cmdHeader + columnArgs + colClose + valueArgs + cmdClose;

                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Info("InsertRecord:2: RunSqlNonQuery returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("InsertRecord: SQL INSERT command failed");
                    return IdLibrary.TriResult.FAIL;
                }

                // Add each class as a call for another table
                logger.Info("InsertRecord:3: Calling mComputers.InsertRecord(ComputerList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = mComputers.InsertRecord("ComputerList", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("InsertRecord: "+mComputers+".InsertRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("InsertRecord: "+mComputers+".InsertRecord() returned "+tResult.ToString());

                logger.Info("InsertRecord:3: Calling mPhones.InsertRecord(PhoneList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = mPhones.InsertRecord("PhoneList", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("InsertRecord: "+mPhones+".InsertRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("InsertRecord: "+mPhones+".InsertRecord() returned "+tResult.ToString());

                logger.Info("InsertRecord:3: Calling mOccupant.InsertRecord(PersonItem, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = mOccupant.InsertRecord("PersonItem", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("InsertRecord: "+mOccupant+".InsertRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("InsertRecord: "+mOccupant+".InsertRecord() returned "+tResult.ToString());

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("InsertRecord: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Update Routines

        /// <summary>
        /// Update all fields of this record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName)
        {
            return UpdateRecordByLowerId(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Update all fields of this record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("UpdateRecordByLowerId: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return UpdateRecordByLowerId(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Update all fields of this record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        /// <param name="aTrans">Transaction to use</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("UpdateRecordByLowerId: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("UpdateRecordByLowerId: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return UpdateRecordByLowerId(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Update all fields of this record in database, using provided connection info.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aDbName">Database Name to use</param>
        /// <param name="aUserID">UserID to log into DB</param>
        /// <param name="aPassword">Password to log into DB</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return UpdateRecordByLowerId(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Update all fields of this record in database. Should only be called by ICDG routines
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string cmdHeader = "UPDATE "+aTableName+" SET ";
                string columnArgs = "UpperID = '"+mUpperID+"', LowerID = '"+mLowerID+"', RoomNumber = '"+mRoomNumber+"'";
                string colClose = " WHERE LowerID =";
                string valueArgs = "'"+mLowerID+"'";
                string cmdClose = ";";
                string command = cmdHeader + columnArgs + colClose + valueArgs + cmdClose;

                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Trace("UpdateRecordByLowerId: RunSqlNonQuery returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("UpdateRecordByLowerId: SQL UPDATE command failed");
                    return IdLibrary.TriResult.FAIL;
                }

                // Add each class as a call for another table
                logger.Info("UpdateRecordByLowerId:3: Calling mComputers.UpdateRecordByLowerId(ComputerList, aConnection, aTrans)");
                tResult = mComputers.UpdateRecordByLowerId("ComputerList", aConnection, aTrans);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("UpdateRecordByLowerId: "+mComputers+".UpdateRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("UpdateRecordByLowerId: "+mComputers+".UpdateRecord() returned "+tResult.ToString());

                logger.Info("UpdateRecordByLowerId:3: Calling mPhones.UpdateRecordByLowerId(PhoneList, aConnection, aTrans)");
                tResult = mPhones.UpdateRecordByLowerId("PhoneList", aConnection, aTrans);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("UpdateRecordByLowerId: "+mPhones+".UpdateRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("UpdateRecordByLowerId: "+mPhones+".UpdateRecord() returned "+tResult.ToString());

                logger.Info("UpdateRecordByLowerId:3: Calling mOccupant.UpdateRecordByLowerId(PersonItem, aConnection, aTrans)");
                tResult = mOccupant.UpdateRecordByLowerId("PersonItem", aConnection, aTrans);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("UpdateRecordByLowerId: "+mOccupant+".UpdateRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("UpdateRecordByLowerId: "+mOccupant+".UpdateRecord() returned "+tResult.ToString());

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("UpdateRecordByLowerId: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Delete Routines

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, string aWhereClause)
        {
            return DeleteRecordByWhere(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, aWhereClause);
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            if (aConnection == null)
            {
                logger.Error("DeleteRecordByWhere: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DeleteRecordByWhere(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, aWhereClause);
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aWhereClause)
        {
            if (aConnection == null)
            {
                logger.Error("DeleteRecordByWhere: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("DeleteRecordByWhere: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DeleteRecordByWhere(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, aWhereClause);
        }

        /// <summary>
        /// Delete this record from database using supplied connection credentials.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, string aDbName, string aUserID, string aPassword, string aWhereClause)
        {
            return DeleteRecordByWhere(aTableName, null, null, aDbName, aUserID, aPassword, aWhereClause);
        }

        /// <summary>
        /// Delete this record from database. Should only be called by IDCG.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword, string aWhereClause)
        {
            IdLibrary.TriResult tResult;

            try
            {
                // Add each class as a call for another table
                logger.Info("DeleteRecordByWhere:3: Calling mComputers.DeleteRecordByLowerId(ComputerList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = mComputers.DeleteRecordByLowerId("ComputerList", aConnection, aTrans, aDbName, aUserID, aPassword);

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DeleteRecordByWhere: "+mComputers+".DeleteRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DeleteRecordByWhere: "+mComputers+".DeleteRecord() returned "+tResult);

                logger.Info("DeleteRecordByWhere:3: Calling mPhones.DeleteRecordByLowerId(PhoneList, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = mPhones.DeleteRecordByLowerId("PhoneList", aConnection, aTrans, aDbName, aUserID, aPassword);

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DeleteRecordByWhere: "+mPhones+".DeleteRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DeleteRecordByWhere: "+mPhones+".DeleteRecord() returned "+tResult);

                logger.Info("DeleteRecordByWhere:3: Calling mOccupant.DeleteRecordByLowerId(PersonItem, aConnection, aTrans, aDbName, aUserID, aPassword)");
                tResult = mOccupant.DeleteRecordByLowerId("PersonItem", aConnection, aTrans, aDbName, aUserID, aPassword);

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DeleteRecordByWhere: "+mOccupant+".DeleteRecord() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DeleteRecordByWhere: "+mOccupant+".DeleteRecord() returned "+tResult);

                string cmdHeader = "DELETE FROM "+aTableName+" WHERE ";
                string valueArgs = aWhereClause;
                string cmdClose = ";";
                string command = cmdHeader + valueArgs + cmdClose;

                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DeleteRecordByWhere: RunSqlNonQuery failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DeleteRecordByWhere: RunSql returned "+tResult.ToString());
                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("DeleteRecordByWhere: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName)
        {
            return DeleteRecordByWhere(aTableName, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, SqlConnection aConnection)
        {
            return DeleteRecordByWhere(aTableName, aConnection, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use</param>
        /// <param name="aTrans">Transaction to use</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            return DeleteRecordByWhere(aTableName, aConnection, aTrans, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aDbName">Database Name to connect to</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return DeleteRecordByWhere(aTableName, aDbName, aUserID, aPassword, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete this record from database. Should only be called by IDCG
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            return DeleteRecordByWhere(aTableName, aConnection, aTrans, aDbName, aUserID, aPassword, "LowerID = '"+ mLowerID + "'");
        }

        #endregion

        #region DB Record Select Open/Read Routines

        /// <summary>
        /// Create and open the SQL select to get fields of this record.
        /// Return reader on success, null on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public SqlDataReader SqlOpen(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            SqlDataReader myReader = null;

            try
            {
                string cmdHeader = "SELECT ";
                string columnArgs = "UpperID, LowerID, RoomNumber";
                string colClose = " FROM "+aTableName+" WHERE " + aWhereClause + ";";
                string command = cmdHeader + columnArgs + colClose;

                logger.Info("SqlOpen: command=["+command+"]");
                SqlCommand myCommand = new SqlCommand(command, aConnection);
                myReader = myCommand.ExecuteReader();
                return myReader;
            }
            catch (Exception ex)
            {
                logger.Error("SqlOpen: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return null;
            }
        }

        /// <summary>
        /// Read in the next sql record into this Item.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aReader">Reader to read from</param>
        public IdLibrary.TriResult SqlRead(SqlDataReader myReader)
        {
            IdLibrary.TriResult tResult = IdLibrary.TriResult.SUCCESS;
            bool result;

            try
            {
                result = myReader.Read();
                if (result == false)
                    return IdLibrary.TriResult.NOTFOUND;

                mUpperID = myReader.GetGuid(0).ToString();
                mLowerID = myReader.GetGuid(1).ToString();
                mComputers = new ComputerList();
                mPhones = new PhoneList();
                mOccupant = new PersonItem();
                mRoomNumber = myReader.GetString(2);

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("SqlRead: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Select Routines

        /// <summary>
        /// Select 1 record from database according to LowerID of this record.
        /// Return SUCCESS, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult SqlSelectByLowerId(string aTableName, SqlConnection aConnection)
        {
            return SqlSelectByWhere(aTableName, aConnection, "LowerID = '"+mLowerID+"'");
        }

        /// <summary>
        /// Select 1 record from database according to given where clause.
        /// Return SUCCESS, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult SqlSelectByWhere(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            SqlDataReader myReader = null;
            IdLibrary.TriResult tResult;

            try
            {
                // Create and open the appropriate select
                myReader = SqlOpen(aTableName, aConnection, aWhereClause);
                if (myReader == null)
                    return IdLibrary.TriResult.FAIL;

                // Since this is an item, get exactly 1 record
                tResult = SqlRead(myReader);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                if (tResult != IdLibrary.TriResult.SUCCESS)
                    return tResult;

                return SqlSelectSubrecords(aConnection);
            }
            catch (Exception ex)
            {
                logger.Error("SqlSelectByWhere: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return IdLibrary.TriResult.FAIL;
            }
        }

        /// <summary>
        /// Load subclasses with their sql data.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        public IdLibrary.TriResult SqlSelectSubrecords(SqlConnection aConnection)
        {
            try
            {
                IdLibrary.TriResult tResult = IdLibrary.TriResult.SUCCESS;

                // Add each class as a call for another table
                string whereClause = "UpperID = '"+mLowerID+"'";
                logger.Info("SqlSelectSubrecords:3: Calling mComputers.SqlSelectByWhere(ComputerList, aConnection, whereClause)");
                tResult = mComputers.SqlSelectByWhere("ComputerList", aConnection, whereClause);

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("SqlSelectSubrecords: mComputers.SqlSelectByWhere() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("SqlSelectSubrecords: mComputers.SqlSelectByWhere() returned "+tResult);

                logger.Info("SqlSelectSubrecords:3: Calling mPhones.SqlSelectByWhere(PhoneList, aConnection, whereClause)");
                tResult = mPhones.SqlSelectByWhere("PhoneList", aConnection, whereClause);

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("SqlSelectSubrecords: mPhones.SqlSelectByWhere() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("SqlSelectSubrecords: mPhones.SqlSelectByWhere() returned "+tResult);

                logger.Info("SqlSelectSubrecords:3: Calling mOccupant.SqlSelectByWhere(PersonItem, aConnection, whereClause)");
                tResult = mOccupant.SqlSelectByWhere("PersonItem", aConnection, whereClause);

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("SqlSelectSubrecords: mOccupant.SqlSelectByWhere() failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("SqlSelectSubrecords: mOccupant.SqlSelectByWhere() returned "+tResult);

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("SqlSelectSubrecords: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #endregion

        #region TestDatabase

        /// <summary>
        /// Basic database test routine for OfficeItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabase(bool aDeleteAtEnd, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult, readResult;
            SqlConnection myConnection = null;
            SqlTransaction myTrans = null;
            OfficeItem fi1, readItem;
            string whereClause;
            string upperID = Guid.NewGuid().ToString();

            logger.Info("");
            logger.Info("TestDatabase:1:");

            // Test building sample record
            try
            {
                fi1 = BuildSample(upperID, 1);
                fi1.Logit(LogLevel.Info, "", "TestDatabase: Sample record:", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: Build sample record test got ex: " + ex);
                return false;
            }

            // Test database open
            logger.Info("TestDatabase:2: Testing OpenDb");
            try
            {
                myConnection = IdLibrary.OpenDbConnection(null, aDbName, aUserID, aPassword);
                if (myConnection == null)
                {
                    logger.Error("TestDatabase: Can't open connection to database '"+aDbName+"'");
                    return false;
                }
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: Open test got ex: " + ex);
                return false;
            }

            // Test drop existing table, ignore table not present
            logger.Info("TestDatabase:3: Testing initial DropTable");
            try
            {
                // Begin transaction
                myTrans = myConnection.BeginTransaction();

                tResult = OfficeItem.DropTable("OfficeItem", myConnection, myTrans);
                if (IdLibrary.AnyIdError(tResult))
                {
                    logger.Info("TestDatabase: Initial drop of 'OfficeItem' returned "+tResult+", ignore");
                }
                else
                {
                    logger.Info("TestDatabase: Initial drop of 'OfficeItem' returned "+tResult);
                }

                // Commit on success
                myTrans.Commit(); myTrans = null;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: DropTable test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            // Test create new table, will fail if table cannot be created
            logger.Info("TestDatabase:4: Testing CreateTable");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = OfficeItem.CreateTable("OfficeItem", myConnection, myTrans);
                logger.Info("TestDatabase: CreateTable test returned " + tResult.ToString());

                if (IdLibrary.AnyIdError(tResult))
                {
                    logger.Info("TestDatabase: Create of 'OfficeItem' failed (Cnt="+tResult.ToString()+")");
                    myTrans.Rollback(); myTrans = null;
                    myConnection.Close();
                    return false;
                }
                else
                {
                    logger.Info("TestDatabase: Create of 'OfficeItem' success");
                }

                // Commit on success
                myTrans.Commit(); myTrans = null;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: CreateTable test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            // Test inserting record into table, will fail if insert fails
            logger.Info("TestDatabase:5: Testing InsertRecord");
            try
            {
                readItem = new OfficeItem();
                readItem.mLowerID = fi1.mLowerID;
                readResult = readItem.SqlSelectByLowerId("OfficeItem", myConnection);
                logger.Info("TestDatabase:5.1: SqlSelectByLowerId returned " + readResult.ToString());

                if (readResult != IdLibrary.TriResult.NOTFOUND)
                {
                    logger.Info("TestDatabase:5.2: Already a "+fi1.mLowerID+" record in db. Insert will fail");
                    myConnection.Close();
                    return false;
                }

                myTrans = myConnection.BeginTransaction();
                tResult = fi1.InsertRecord("OfficeItem", myConnection, myTrans);
                if (IdLibrary.AnyIdError(tResult))
                {
                    logger.Info("TestDatabase: Insert record 1 for 'OfficeItem' failed (Cnt="+tResult.ToString()+")");
                    myTrans.Rollback(); myTrans = null;
                    myConnection.Close();
                    return false;
                }

                myTrans.Commit(); myTrans = null;
                readItem = new OfficeItem();
                readItem.mLowerID = fi1.mLowerID;
                readResult = readItem.SqlSelectByLowerId("OfficeItem", myConnection);

                if (readResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabase:5.4: Failed to get "+fi1.mLowerID+" record from db.");
                    myConnection.Close();
                    return false;
                }

                readItem.Logit(LogLevel.Info, "", "TestDatabase: After insert:", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: InsertRecord test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            // Test updating record into table, will fail if insert fails
            logger.Info("TestDatabase:6: Testing UpdateRecordbyLowerId");
            try
            {
                whereClause = "LowerID = '"+fi1.mLowerID+"'";

                if (fi1.UpdateSample() == false)
                {
                    logger.Info("TestDatabase: Update Sample for 'OfficeItem' failed ");
                    myConnection.Close();
                    return false;
                }

                fi1.Logit(LogLevel.Info, "", "TestDatabase: After update sample:", IdLibrary.LogDocDepth.MAX);
                myTrans = myConnection.BeginTransaction();

                tResult = fi1.UpdateRecordByLowerId("OfficeItem", myConnection, myTrans);
                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabase: Update record 1 for 'OfficeItem' failed (Cnt="+tResult.ToString()+")");
                    myTrans.Rollback();
                    myConnection.Close();
                    return false;
                }

                myTrans.Commit(); myTrans = null;
                readItem = new OfficeItem();
                readItem.mLowerID = fi1.mLowerID;
                readResult = readItem.SqlSelectByWhere("OfficeItem", myConnection, whereClause);

                if (readResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabase:6.4: Failed to get "+fi1.mLowerID+" record from db.");
                    myConnection.Close();
                    return false;
                }

                readItem.Logit(LogLevel.Info, "", "TestDatabase: After update:", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: UpdateRecord test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            if (aDeleteAtEnd == true)
            {
                if (TestDatabaseFinalRecordDelete(myConnection, fi1, whereClause) == false)
                {
                    logger.Info("TestDatabase: TestDatabaseFinalRecordDelete() failed");
                    myConnection.Close();
                    return false;
                }

                if (TestDatabaseDropTable(myConnection) == false)
                {
                    logger.Info("TestDatabase: TestDatabaseDropTable() failed");
                    myConnection.Close();
                    return false;
                }
            }

            // Success
            logger.Info("TestDatabase: tests completed successfully");
            myConnection.Close();
            return true;
        }

        /// <summary>
        /// database final table delete test routine for OfficeItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabaseFinalRecordDelete(SqlConnection myConnection, OfficeItem fi1, string whereClause)
        {
            IdLibrary.TriResult tResult;
            OfficeItem readItem;
            SqlTransaction myTrans = null;

            // Test deleting record into table, will fail if insert fails
            logger.Info("TestDatabaseFinalRecordDelete:1: Testing DeleteRecordByWhere");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = fi1.DeleteRecordByWhere("OfficeItem", myConnection, myTrans, whereClause);
                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabaseFinalTableDelete: Delete record for 'OfficeItem' failed (Cnt="+tResult.ToString()+")");
                    if (myTrans != null)
                    {
                        myTrans.Rollback(); myTrans = null;
                    }
                    return false;
                }

                myTrans.Commit(); myTrans = null;
                readItem = new OfficeItem();
                readItem.mLowerID = fi1.mLowerID;
                tResult = readItem.SqlSelectByWhere("OfficeItem", myConnection, whereClause);

                if (tResult != IdLibrary.TriResult.NOTFOUND)
                {
                    readItem.Logit(LogLevel.Error, "", "TestDatabaseFinalRecordDelete: Record was supposed to be deleted:", IdLibrary.LogDocDepth.MAX);
                    return false;
                }

                return true;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabaseFinalRecordDelete: DeleteRecordByWhere test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                return false;
            }
        }

        /// <summary>
        /// database drop test routine for OfficeItem.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabaseDropTable(SqlConnection myConnection)
        {
            IdLibrary.TriResult tResult;
            SqlTransaction myTrans = null;

            // Test drop existing table, will fail if table not present
            logger.Info("TestDatabaseDropTable:1: Testing final DropTable");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = OfficeItem.DropTable("OfficeItem", myConnection, myTrans);
                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabaseDropTable: Final drop of 'OfficeItem' returned "+tResult.ToString()+", returning false");
                    if (myTrans != null)
                    {
                        myTrans.Rollback(); myTrans = null;
                    }
                    return false;
                }

                // Success
                logger.Info("TestDatabaseDropTable: test completed successfully");
                myTrans.Commit(); myTrans = null;
                return true;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabaseDropTable: got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                return false;
            }
        }

        #endregion

        #region Class-Specific

        /// <summary>
        /// Sample function to comment on any Windows system.
        /// This was read in from OfficeItemCode.txt
        /// </summary>
        public void BooWindows()
        {
            mComputers.BooWindows();
        }

        #endregion


        #endregion

        #region Private Access

        #region Private Members

        private string mUpperID;
        private string mLowerID;
        private ComputerList mComputers;
        private PhoneList mPhones;
        private PersonItem mOccupant;
        private string mRoomNumber;

        #endregion

        #endregion
    }
}
