using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using NLog;
using System.IO;
using System.Collections;
using System.Data.SqlClient;

#region Class-Specific Usings


#endregion

namespace IdcgLibrary
{
    [Serializable]
    public class PhoneList
    {
        #region Constants

        public static string CLASS_NAME = typeof(PhoneList).ToString();

        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        #endregion

        #region Public Access

        #region Public Properties

        public string UpperIDValue
        {
            get { return mUpperID; }
            set { mUpperID = value; }
        }

        public string LowerIDValue
        {
            get { return mLowerID; }
            set { mLowerID = value; }
        }

        public List<PhoneItem> TheListValue
        {
            get { return mTheList; }
            set { mTheList = value; }
        }

        public int CurIdxValue
        {
            get { return mCurIdx; }
            set { mCurIdx = value; }
        }

        [XmlIgnore]
        // Indexer to give array access to the main list.
        public PhoneItem this[int index]
        {
            // Return the item at this index
            get
            {
                if (index < 0 || index >= mTheList.Count)
                    return null;

                return mTheList[index];
            }
            // Set the item at this index to given value
            set
            {
                if (index < 0 || index >= mTheList.Count)
                    return;

                mTheList[index] = value;
            }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Set default values for all fields.
        /// </summary>
        public PhoneList()
        {
            Reset();
        }

        /// <summary>
        /// Set all fields according to arguments
        /// </summary>
        /// <param name=""></param>
        public PhoneList(string aUpperID, string aLowerID, List<PhoneItem> aTheList, int aCurIdx)
        {
            Set(aUpperID, aLowerID, aTheList, aCurIdx);
        }

        /// <summary>
        /// Copy Constructor
        /// </summary>
        /// <param name=""></param>
        public PhoneList(PhoneList aItem)
        {
            Set(aItem);
        }

        #endregion

        #region Reset, Set

        /// <summary>
        /// Set fields to default values
        /// </summary>
        public void Reset()
        {
            mUpperID = "00000000-0000-0000-0000-000000000000";
            mLowerID = "00000000-0000-0000-0000-000000000000";
            mTheList = new List<PhoneItem>();
            mCurIdx = -1;
        }

        /// <summary>
        /// Set the values of every field according to arguments
        /// </summary>
        /// <param name=""></param>
        public void Set(string aUpperID, string aLowerID, List<PhoneItem> aTheList, int aCurIdx)
        {
            mUpperID = aUpperID;
            mLowerID = aLowerID;
            
            mTheList = new List<PhoneItem>();
            foreach (PhoneItem item in aTheList)
            {
                mTheList.Add(new PhoneItem(item));
            }

            mCurIdx = aCurIdx;
        }

        /// <summary>
        /// Copy our values from given PhoneList
        /// </summary>
        /// <param name="aFrom"></param>
        public void Set(PhoneList aFrom)
        {
            mUpperID = aFrom.mUpperID;
            mLowerID = aFrom.mLowerID;
            
            mTheList = new List<PhoneItem>();
            foreach (PhoneItem item in aFrom.mTheList)
            {
                mTheList.Add(new PhoneItem(item));
            }

            mCurIdx = aFrom.mCurIdx;
        }

        #endregion

        #region Standard Methods

        /// <summary>
        /// Add the given item to the end of the list.
        /// </summary>
        /// <param name="item"></param>
        public void AddItem(PhoneItem item)
        {
            PhoneItem tempItem = new PhoneItem(item);
            tempItem.UpperIDValue = mLowerID;
            mTheList.Add(tempItem);
        }

        /// <summary>
        /// Return the number of items in the list.
        /// </summary>
        /// <returns></returns>
        public int GetCount()
        {
            return mTheList.Count;
        }

        /// <summary>
        /// Return the item at the requested index.
        /// </summary>
        /// <param name="idx"></param>
        /// <returns></returns>
        public PhoneItem GetItem(int idx)
        {
            if (idx >= mTheList.Count)
                return null;

            return mTheList[idx];
        }

        /// <summary>
        /// Remove the item at the given index
        /// </summary>
        /// <param name="aIdx"></param>
        /// <returns></returns>
        public bool RemoveAt(int aIdx)
        {
            if (aIdx < 0 || aIdx >= mTheList.Count)
                return false;

            mTheList.RemoveAt(aIdx);
            return true;
        }

        /// <summary>
        /// Allows foreach to work with our list
        /// </summary>
        /// <returns></returns>
        public IEnumerator GetEnumerator()
        {
            foreach (object item in mTheList)
            {
                yield return item;
            }
        }

        #endregion

        #region Find / Get Methods

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aUpperID"></param>
        /// <returns></returns>
        public int FindUpperID(string aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                PhoneItem item = mTheList[i];

                if (item.UpperIDValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aUpperID"></param>
        /// <returns></returns>
        public PhoneItem GetUpperID(string aValue)
        {
            foreach (PhoneItem item in mTheList)
            {
                if (item.UpperIDValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aLowerID"></param>
        /// <returns></returns>
        public int FindLowerID(string aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                PhoneItem item = mTheList[i];

                if (item.LowerIDValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aLowerID"></param>
        /// <returns></returns>
        public PhoneItem GetLowerID(string aValue)
        {
            foreach (PhoneItem item in mTheList)
            {
                if (item.LowerIDValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aManufacturer"></param>
        /// <returns></returns>
        public int FindManufacturer(string aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                PhoneItem item = mTheList[i];

                if (item.ManufacturerValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aManufacturer"></param>
        /// <returns></returns>
        public PhoneItem GetManufacturer(string aValue)
        {
            foreach (PhoneItem item in mTheList)
            {
                if (item.ManufacturerValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aSerialNumber"></param>
        /// <returns></returns>
        public int FindSerialNumber(string aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                PhoneItem item = mTheList[i];

                if (item.SerialNumberValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aSerialNumber"></param>
        /// <returns></returns>
        public PhoneItem GetSerialNumber(string aValue)
        {
            foreach (PhoneItem item in mTheList)
            {
                if (item.SerialNumberValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aIsDigital"></param>
        /// <returns></returns>
        public int FindIsDigital(bool aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                PhoneItem item = mTheList[i];

                if (item.IsDigitalValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aIsDigital"></param>
        /// <returns></returns>
        public PhoneItem GetIsDigital(bool aValue)
        {
            foreach (PhoneItem item in mTheList)
            {
                if (item.IsDigitalValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aLastChange"></param>
        /// <returns></returns>
        public int FindLastChange(DateTime aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                PhoneItem item = mTheList[i];

                if (item.LastChangeValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aLastChange"></param>
        /// <returns></returns>
        public PhoneItem GetLastChange(DateTime aValue)
        {
            foreach (PhoneItem item in mTheList)
            {
                if (item.LastChangeValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Return item at the current index, or null if none
        /// </summary>
        /// <returns></returns>
        public PhoneItem GetCurrentItem()
        {
            if (mCurIdx < 0 || mCurIdx >= mTheList.Count)
                return null;

            return mTheList[mCurIdx];
        }

        #endregion

        #region Logging

        /// <summary>
        /// Return ToString for this class, or "null"
        /// </summary>
        /// <param name="aValue"></param>
        /// <returns></returns>
        public static string ToStringP(PhoneList aValue)
        {
            if (aValue == null)
                return "null";
            else
                return aValue.ToString();
        }

        /// <summary>
        /// Return fields of this object as a string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "mUpperID="+mUpperID+", mLowerID="+mLowerID+", mTheList= "+((mTheList == null) ? "null" : "(Cnt="+mTheList.Count+")")+", mCurIdx="+mCurIdx;
        }

        /// <summary>
        /// Log this object to nlog
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void Logit(LogLevel aLevel, string aIndent, string aHeader, IdLibrary.LogDocDepth aDepth)
        {
            string logString = LogitDetailToString(aLevel, aIndent, aHeader, aDepth);
            logger.Log(aLevel, logString);
        }


        /// <summary>
        /// Log the contents of this class to String at various levels of detail
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitDetailToString(LogLevel aLevel, string aIndent, string aHeader, IdLibrary.LogDocDepth aDepth)
        {
            #region Init

            // Don't log if not within level
            if (IdLibrary.WithinLoggerLevel(logger, aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            #endregion

            // Minimum level does a one-line summary
            if (aDepth == IdLibrary.LogDocDepth.MINIMUM)
            {
                sb.Append(aIndent + aHeader + " PhoneList. CurIdx=" + mCurIdx + ", item count=" + mTheList.Count + ":" );

                return sb.ToString();
            }

            // Any level above MINIMUM
            sb.AppendLine();
            sb.AppendLine(aIndent + aHeader + " PhoneList. CurIdx=" + mCurIdx + ", item count=" + mTheList.Count + " (" + aDepth + "):");

            int childIdx;

            for (childIdx = 0; childIdx < mTheList.Count; childIdx++)
            {
                string docValue;

                switch (aDepth)
                {
                     case IdLibrary.LogDocDepth.MAX:
                         docValue = mTheList[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:", aDepth);
                         break;

                     case IdLibrary.LogDocDepth.FULL:
                         docValue = mTheList[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:",
                                        IdLibrary.LogDocDepth.MODERATE);
                         break;

                     case IdLibrary.LogDocDepth.MODERATE:
                         docValue = mTheList[childIdx].LogitDetailToString(aLevel, aIndent + "    ", "[" + childIdx + "]:",
                                        IdLibrary.LogDocDepth.MINIMUM);
                         break;

                     default:
                         docValue = "unknown";   // should not happen
                         break;
                }

                sb.AppendLine(docValue);
            }

            sb.AppendLine(aIndent + "        "+"mUpperID="+mUpperID+", mLowerID="+mLowerID);

            sb.Append(aIndent + "    End of PhoneList");
            return sb.ToString();
        }


        #endregion

        #region I/O Routines

        /// <summary>
        /// Convert our class to an XML string via Serialize. Can throw exceptions
        /// </summary>
        /// <returns></returns>
        public string ToXML()
        {
            try
            {
                // Force UTF-8
                XmlSerializer serializer = new XmlSerializer(typeof(PhoneList));
                using (StringWriter writer = new IdLibrary.Utf8StringWriter())
                {
                    serializer.Serialize(writer, this);
                    return writer.ToString();
                }
            }
            catch (Exception ex)
            {
                logger.Error("ToXML: Exception '" + ex + "'");
                throw new Exception("ToXML", ex);
            }
        }

        /// <summary>
        /// Convert given xml string to our class. Throw exception on fail.
        /// </summary>
        /// <param name="xml"></param>
        /// <returns></returns>
        public PhoneList FromXML(string xml)
        {
            PhoneList ei = null;
            if ((xml != null) && (xml.Length > 0))
            {
                try
                {
                    XmlSerializer serializer = new XmlSerializer(typeof(PhoneList));
                    ei = (PhoneList)serializer.Deserialize(new StringReader(xml));
                }
                catch (Exception ex)
                {
                    logger.Error("FromXML: Got exception '" + ex + "'");
                    throw new Exception("FromXML", ex);
                }
            }
            return ei;
        }

        /// <summary>
        /// Write out our class to given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool WriteFile(string fileName)
        {
            try
            {
                StreamWriter writer = new StreamWriter(fileName);
                writer.Write(this.ToXML());
                writer.Close();
            }
            catch (Exception ex)
            {
                logger.Info("WriteFile: Got exception '" + ex + "'");
                return false;
            }
            // Success
            return true;
        }

        /// <summary>
        /// Read our class from given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool ReadFile(string fileName)
        {
            try
            {
                StreamReader reader = new StreamReader(fileName);
                string contents = reader.ReadToEnd();
                reader.Close();

                // Set all our values from what we got back
                PhoneList fi = FromXML(contents);
                this.Set(fi);
            }
            catch (Exception ex)
            {
                logger.Info("ReadFile: Got exception '" + ex + "'");
                return false;
            }

            // Success
            return true;
        }

        #endregion

        #region Tester

        /// <summary>
        /// Basic test routine item and list.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool Tester()
        {
            string TEST_FILE = "testPhone.xml";
            PhoneItem item1, item2, item3;
            PhoneList list1, list2;
            int findIdx;
            bool result;
            string upperID;
            string listLowerID;

            // Test creating the items and building a list
            try
            {
                logger.Info("");

                // Create list #1
                upperID = Guid.NewGuid().ToString();
                list1 = new PhoneList("24c286ae-b964-485e-bd31-5bd10a3e32b1", "fea990b0-a3f0-451f-b213-9d58067d139a", new List<PhoneItem>(), 2);
                listLowerID = list1.LowerIDValue;
                item1 = PhoneItem.BuildSample(listLowerID, 1);
                list1.AddItem(item1);
                logger.Info("Tester:2: 1 item list string=[" + list1.ToString() + "]");

                // Create list #2
                upperID = Guid.NewGuid().ToString();
                list2 = new PhoneList(list1);
                list2.LowerIDValue = Guid.NewGuid().ToString();
                listLowerID = list2.LowerIDValue;

                foreach (PhoneItem updateItem in list2)
                    updateItem.UpperIDValue = listLowerID;

                item2 = PhoneItem.BuildSample(listLowerID, 2);
                list2.AddItem(item2);

                item3 = PhoneItem.BuildSample(listLowerID, 3);
                list2.AddItem(item3);

                list2.Logit(LogLevel.Info, "", "Tester:3: List2:", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Build list section got ex: " + ex);
                return false;
            }

            // Test logging and finds
            try
            {
                list2.Logit(LogLevel.Info, "", "Tester:4: created list:", IdLibrary.LogDocDepth.MAX);

                // Supposed to fail
                findIdx = list2.FindUpperID("00000000-0000-0000-0000-000000000000");

                if (findIdx >= 0)
                {
                    logger.Info("Tester: FindUpperID(00000000-0000-0000-0000-000000000000) succeeded, returning false");
                    return false;
                }

                // Supposed to succeed
                findIdx = list2.FindUpperID(list2[0].UpperIDValue);

                if (findIdx < 0)
                {
                    logger.Info("Tester:5: FindUpperID(list2[0].UpperIDValue) failed, returning false");
                    return false;
                }

                // Test removal
                result = list2.RemoveAt(findIdx);

                if (result != true)
                {
                    logger.Info("Tester:6: list2.RemoveAt(" + findIdx + ") failed, returning false");
                    return false;
                }

                list2.Logit(LogLevel.Info, "", "Tester: After remove, list2:", IdLibrary.LogDocDepth.MINIMUM);
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Build list section got ex: " + ex);
                return false;
            }

            try
            {
                // Write list2 out to XML
                if (list2.WriteFile(TEST_FILE) == false)
                {
                    logger.Info("Tester:7: list2 WriteFile("+TEST_FILE+") failed, returning false");
                    return false;
                }

                // Read in as ul1
                result = list1.ReadFile(TEST_FILE);

                if (result != true)
                {
                    logger.Info("Tester:8: ReadFile(" + TEST_FILE + ") failed, returning false");
                    return false;
                }

                list1.Logit(LogLevel.Info, "", "Tester:9: Read "+TEST_FILE+":", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Read/Write list section got ex: " + ex);
                return false;
            }

            // Success
            return true;
        }

        /// <summary>
        /// Return a sample record PhoneList.
        /// Return sample on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static PhoneList BuildSample(string aUpperID, int aValue)
        {
            PhoneItem item;
            PhoneList rec;

            rec = new PhoneList("cc7f82c7-32b2-4898-836f-6b2a523af339", "6b8ffffd-f6c7-43fb-98b4-346381bd3307", new List<PhoneItem>(), 1);
            rec.UpperIDValue = aUpperID;
            rec.LowerIDValue = Guid.NewGuid().ToString();

            // Build list of 3 of same kind
            if (aValue == 1 || aValue == 2 || aValue == 3)
            {
                item = PhoneItem.BuildSample(rec.LowerIDValue, aValue);
                rec.AddItem(item);
                item = PhoneItem.BuildSample(rec.LowerIDValue, aValue);
                rec.AddItem(item);
                item = PhoneItem.BuildSample(rec.LowerIDValue, aValue);
                rec.AddItem(item);
            }
            else
            {
                item = PhoneItem.BuildSample(rec.LowerIDValue, -1);
                rec.AddItem(item);
                item = PhoneItem.BuildSample(rec.LowerIDValue, -1);
                rec.AddItem(item);
                item = PhoneItem.BuildSample(rec.LowerIDValue, -1);
                rec.AddItem(item);
            }

            return rec;
        }

        /// <summary>
        /// Drill down through sample record until non-class 2nd item and update it
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public bool UpdateSample()
        {
            logger.Trace("UpdateSample:1:");

            foreach (PhoneItem rec in mTheList)
            {
                if (rec.UpdateSample() == false)
                    return false;
            }

            return true;
        }

        #endregion

        #region Database Routines

        #region Create Table Routines

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public static IdLibrary.TriResult CreateTable(string aTableName)
        {
            return CreateTable(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("CreateTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return CreateTable(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Create given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("CreateTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("CreateTable: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return CreateTable(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Create given table in database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return CreateTable(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Create given table in database. This should only be called by IDCG!
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult CreateTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string cmdHeader = "CREATE TABLE "+aTableName+"(";
                string cmdArgs = "UpperID UNIQUEIDENTIFIER DEFAULT NEWID(), LowerID UNIQUEIDENTIFIER DEFAULT NEWID(), CurIdx int";
                string cmdClose = ")";
                string command = cmdHeader + cmdArgs + cmdClose;
                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Trace("CreateTable: RunSql returned "+tResult.ToString());

                // Unfortunately, create table returns -1 on success or fail. Check for exception (-2)
                if (tResult == IdLibrary.TriResult.FAIL)
                    return tResult;

                // Add each class as a call for another table
                tResult = PhoneItem.CreateTable("PhoneItem", aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                    return IdLibrary.TriResult.FAIL;

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("CreateTable: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region Drop Table Routines

        /// <summary>
        /// Drop given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to Drop</param>
        public static IdLibrary.TriResult DropTable(string aTableName)
        {
            return DropTable(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Drop given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to Drop</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult DropTable(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("DropTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DropTable(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Drop given table in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to Drop</param>
        /// <param name="aConnection">Connection to use</param>
        public static IdLibrary.TriResult DropTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("DropTable: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("DropTable: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }
            return DropTable(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Drop given table in database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to Drop</param>
        public static IdLibrary.TriResult DropTable(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return DropTable(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Destroy given table in database. This should only be called by IDCG!
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public static IdLibrary.TriResult DropTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string command = "DROP TABLE dbo."+aTableName;
                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Trace("DropTable: RunSql returned "+tResult.ToString());

                // Unfortunately, drop table returns -1 on success or fail. Check for exception (-2). NotFound(-3) is okay
                if (tResult == IdLibrary.TriResult.FAIL)
                    return tResult;

                // Add each class as a call for another table
                tResult = PhoneItem.DropTable("PhoneItem", aConnection, aTrans, aDbName, aUserID, aPassword);

                if (tResult == IdLibrary.TriResult.FAIL)
                    return IdLibrary.TriResult.FAIL;

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("DropTable: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Insert Routines

        /// <summary>
        /// Insert record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        public IdLibrary.TriResult InsertRecord(string aTableName)
        {
            return InsertRecord(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Insert record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("InsertRecord: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return InsertRecord(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Insert record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to insert into</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("InsertRecord: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("InsertRecord: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return InsertRecord(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Insert record into database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return InsertRecord(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Insert list records into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult InsertRecord(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string cmdHeader = "INSERT INTO "+aTableName+" (";
                string columnArgs = "UpperID, LowerID, CurIdx";
                string colClose = ") Values (";
                string valueArgs = "'"+mUpperID+"', '"+mLowerID+"', "+mCurIdx+"";
                string cmdClose = ");";
                string command = cmdHeader + columnArgs + colClose + valueArgs + cmdClose;

                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Info("InsertRecord:2: RunSqlNonQuery returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("InsertRecord: SQL INSERT command failed");
                    return IdLibrary.TriResult.FAIL;
                }

                // Add each class as a call for another table
                foreach (PhoneItem item in mTheList)
                {
                    tResult = item.InsertRecord("PhoneItem", aConnection, aTrans, aDbName, aUserID, aPassword);

                    if (tResult == IdLibrary.TriResult.FAIL)
                    {
                        item.Logit(LogLevel.Error, "", "item.InsertRecord: Failed to insert:", IdLibrary.LogDocDepth.MAX);
                        return IdLibrary.TriResult.FAIL;
                    }
                }

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("InsertRecord: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Update Routines

        /// <summary>
        /// Update record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to update into</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName)
        {
            return UpdateRecordByLowerId(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Update record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to update into</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("UpdateRecord: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return UpdateRecordByLowerId(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Update record into database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to update into</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("UpdateRecordByLowerId: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("UpdateRecordByLowerId: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return UpdateRecordByLowerId(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD);
        }

        /// <summary>
        /// Update record into database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to update</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return UpdateRecordByLowerId(aTableName, null, null, aDbName, aUserID, aPassword);
        }

        /// <summary>
        /// Update all fields of all list records in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult UpdateRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult;

            try
            {
                string cmdHeader = "UPDATE "+aTableName+" SET ";
                string columnArgs = "UpperID = '"+mUpperID+"', LowerID = '"+mLowerID+"', CurIdx = "+mCurIdx+"";
                string colClose = " WHERE LowerID =";
                string valueArgs = "'"+mLowerID+"'";
                string cmdClose = ";";
                string command = cmdHeader + columnArgs + colClose + valueArgs + cmdClose;

                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                logger.Info("UpdateRecordByLowerId:2: RunSqlNonQuery returned "+tResult.ToString());

                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("UpdateRecordByLowerId: SQL UPDATE command failed");
                    return IdLibrary.TriResult.FAIL;
                }

                // Add each class as a call for another table
                foreach (PhoneItem item in mTheList)
                {
                    tResult = item.UpdateRecordByLowerId("PhoneItem", aConnection, aTrans, aDbName, aUserID, aPassword);

                    if (tResult == IdLibrary.TriResult.FAIL)
                    {
                        item.Logit(LogLevel.Error, "", "item.UpdateRecordByLowerId: Failed to insert:", IdLibrary.LogDocDepth.MAX);
                        return IdLibrary.TriResult.FAIL;
                    }
                }

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("UpdateRecordByLowerId: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Delete Routines

        /// <summary>
        /// Delete record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete from</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName)
        {
            return DeleteRecordByWhere(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete from</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, SqlConnection aConnection)
        {
            if (aConnection == null)
            {
                logger.Error("DeleteRecordByLowerId: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DeleteRecordByWhere(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete from</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            if (aConnection == null)
            {
                logger.Error("DeleteRecordByLowerId: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("DeleteRecordByLowerId: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DeleteRecordByWhere(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete record from database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete from</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, string aDbName, string aUserID, string aPassword)
        {
            return DeleteRecordByWhere(aTableName, null, null, aDbName, aUserID, aPassword, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete record from database. This should only be called by IDCG!
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete from</param>
        public IdLibrary.TriResult DeleteRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword)
        {
            return DeleteRecordByWhere(aTableName, aConnection, aTrans, aDbName, aUserID, aPassword, "LowerID = '"+ mLowerID + "'");
        }

        /// <summary>
        /// Delete record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete from</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, string aWhereClause)
        {
            return DeleteRecordByWhere(aTableName, null, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, aWhereClause);
        }

        /// <summary>
        /// Delete record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete from</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            if (aConnection == null)
            {
                logger.Error("DeleteRecord: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DeleteRecordByWhere(aTableName, aConnection, null, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, aWhereClause);
        }

        /// <summary>
        /// Delete record in database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete from</param>
        /// <param name="aConnection">Connection to use</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aWhereClause)
        {
            if (aConnection == null)
            {
                logger.Error("DeleteRecord: Null connection not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            if (aTrans == null)
            {
                logger.Error("DeleteRecord: Null transaction not allowed");
                return IdLibrary.TriResult.FAIL;
            }

            return DeleteRecordByWhere(aTableName, aConnection, aTrans, IdLibrary.DEFT_DBNAME, IdLibrary.DEFT_USERID, IdLibrary.DEFT_USERPASSWORD, aWhereClause);
        }

        /// <summary>
        /// Delete record from database, using connection info given.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to delete from</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, string aDbName, string aUserID, string aPassword, string aWhereClause)
        {
            return DeleteRecordByWhere(aTableName, null, null, aDbName, aUserID, aPassword, aWhereClause);
        }

        /// <summary>
        /// Delete this record from database.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult DeleteRecordByWhere(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aDbName, string aUserID, string aPassword, string aWhereClause)
        {
            IdLibrary.TriResult tResult;

            try
            {
                // Add each class as a call for another table
                foreach (PhoneItem item in mTheList)
                {
                    tResult = item.DeleteRecordByLowerId("PhoneItem", aConnection, aTrans, aDbName, aUserID, aPassword);

                    if (tResult == IdLibrary.TriResult.FAIL)
                    {
                        item.Logit(LogLevel.Error, "", "item.DeleteRecordByLowerId: Failed to delete:", IdLibrary.LogDocDepth.MAX);
                        return IdLibrary.TriResult.FAIL;
                    }
                }
                string cmdHeader = "DELETE FROM "+aTableName+" WHERE ";
                string valueArgs = aWhereClause;
                string cmdClose = ";";
                string command = cmdHeader + valueArgs + cmdClose;

                tResult = IdLibrary.RunSqlNonQuery(command, aConnection, aTrans, aDbName, aUserID, aPassword);
                if (tResult == IdLibrary.TriResult.FAIL)
                {
                    logger.Error("DeleteRecordByWhere: RunSqlNonQuery failed");
                    return IdLibrary.TriResult.FAIL;
                }

                logger.Trace("DeleteRecordByWhere: RunSql returned "+tResult.ToString());
                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("DeleteRecordByWhere: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Select Open/Read Routines

        /// <summary>
        /// Create and open the SQL select to get fields of this record.
        /// Return reader on success, null on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public SqlDataReader SqlOpen(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            SqlDataReader myReader = null;

            try
            {
                string cmdHeader = "SELECT ";
                string columnArgs = "UpperID, LowerID, CurIdx";
                string colClose = " FROM "+aTableName+" WHERE " + aWhereClause + ";";
                string command = cmdHeader + columnArgs + colClose;

                logger.Trace("SqlOpen: command=["+command+"]");
                SqlCommand myCommand = new SqlCommand(command, aConnection);
                myReader = myCommand.ExecuteReader();
                return myReader;
            }
            catch (Exception ex)
            {
                logger.Error("SqlOpen: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return null;
            }
        }

        /// <summary>
        /// Read in the next sql record into this Item.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aReader">Reader to read from</param>
        public IdLibrary.TriResult SqlRead(SqlDataReader myReader)
        {
            IdLibrary.TriResult tResult = IdLibrary.TriResult.SUCCESS;
            bool result;

            try
            {
                result = myReader.Read();
                if (result == false)
                    return IdLibrary.TriResult.NOTFOUND;

                mUpperID = myReader.GetGuid(0).ToString();
                mLowerID = myReader.GetGuid(1).ToString();
                mTheList = new List<PhoneItem>();
                mCurIdx = myReader.GetInt32(2);

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("SqlRead: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #region DB Record Select Routines

        /// <summary>
        /// Select multiple records from database.
        /// Return SUCCESS, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult SqlSelectByLowerId(string aTableName, SqlConnection aConnection)
        {
            return SqlSelectByWhere(aTableName, aConnection, "LowerID = '" + mLowerID+"'");
        }

        /// <summary>
        /// Select multiple records from database.
        /// Return SUCCESS, FAIL, or NOTFOUND
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public IdLibrary.TriResult SqlSelectByWhere(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            SqlDataReader myReader = null;
            IdLibrary.TriResult tResult;

            try
            {
                logger.Trace("SqlSelectByWhere:1: aTableName='"+aTableName+"', aWhereClause='"+aWhereClause+"'");

                // Create and open the appropriate select
                myReader = SqlOpen(aTableName, aConnection, aWhereClause);
                if (myReader == null)
                    return IdLibrary.TriResult.FAIL;

                // Since this is our fields, get exactly 1 record
                tResult = SqlRead(myReader);
                logger.Trace("SqlSelectByWhere:2: SqlRead returned "+tResult.ToString());

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                if (tResult != IdLibrary.TriResult.SUCCESS)
                    return tResult;

                return SqlSelectSubrecords(aConnection);
            }
            catch (Exception ex)
            {
                logger.Error("SqlSelectByWhere: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return IdLibrary.TriResult.FAIL;
            }
        }

        /// <summary>
        /// Load subclasses with their sql data.
        /// Return SUC, FAIL, or NOTFOUND
        /// </summary>
        public IdLibrary.TriResult SqlSelectSubrecords(SqlConnection aConnection)
        {
            SqlDataReader myReader = null;
            IdLibrary.TriResult tResult = IdLibrary.TriResult.SUCCESS;

            try
            {
                // Add each class as a call for another table
                string whereClause = "UpperID = '"+mLowerID+"'";
                logger.Trace("SqlSelectSubrecords:1: whereClause='"+whereClause+"'");

                PhoneItem listItem = new PhoneItem();

                // Create and open the appropriate select
                myReader = listItem.SqlOpen("PhoneItem", aConnection, whereClause);
                if (myReader == null)
                {
                    logger.Info("SqlSelectSubrecords: SqlOpen(PhoneItem) failed");
                    return IdLibrary.TriResult.FAIL;
                }

                // Read until run out of records
                while (true)
                {
                    tResult = listItem.SqlRead(myReader);
                    if (tResult == IdLibrary.TriResult.FAIL)
                    {
                        logger.Info("SqlSelectSubrecords: SqlRead() failed");
                        if (myReader != null && myReader.IsClosed == false)
                            myReader.Close();

                        return IdLibrary.TriResult.FAIL;
                    }
                    else if (tResult == IdLibrary.TriResult.NOTFOUND)
                    {
                       if (mTheList.Count > 0)
                           tResult = IdLibrary.TriResult.SUCCESS;
                       else
                           tResult = IdLibrary.TriResult.NOTFOUND;
                       break;
                    }

                    mTheList.Add(new PhoneItem(listItem));
                    tResult = IdLibrary.TriResult.SUCCESS;
                }

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                foreach (PhoneItem item in mTheList)
                    if (item.SqlSelectSubrecords(aConnection) == IdLibrary.TriResult.FAIL)
                        return IdLibrary.TriResult.FAIL;

                return tResult;
            }
            catch (Exception ex)
            {
                logger.Error("SqlSelectSubrecords: Got ex: "+ex);
                return IdLibrary.TriResult.FAIL;
            }
        }

        #endregion

        #endregion

        #region TestDatabase

        /// <summary>
        /// Basic database test routine for PhoneList.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabase(bool aDeleteAtEnd, string aDbName, string aUserID, string aPassword)
        {
            IdLibrary.TriResult tResult, readResult;
            SqlConnection myConnection;
            SqlTransaction myTrans = null;
            PhoneList sampleList, readList;
            string whereClause;
            string upperID = Guid.NewGuid().ToString();

            logger.Info("");
            logger.Info("TestDatabase:1: upperID="+upperID);

            // Test building sample list
            try
            {
                sampleList = BuildSample(upperID, 1);
                sampleList.Logit(LogLevel.Info, "", "TestDatabase: Sample list:", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: Build sample list test got ex: " + ex);
                return false;
            }

            // Test database open
            logger.Info("TestDatabase:2: Testing OpenDb");
            try
            {
                myConnection = IdLibrary.OpenDbConnection(null, aDbName, aUserID, aPassword);
                if (myConnection == null)
                {
                    logger.Error("TestDatabase: Can't open connection to database '"+aDbName+"'");
                    return false;
                }
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: Open test got ex: " + ex);
                return false;
            }

            // Test drop existing tables, ignore tables not present
            logger.Info("TestDatabase:3: Testing initial DropTable");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = PhoneList.DropTable("PhoneList", myConnection, myTrans);
                if (IdLibrary.AnyIdError(tResult))
                {
                    logger.Info("TestDatabase: Initial drop of 'PhoneList' returned "+tResult.ToString()+", ignore");
                }
                else
                {
                    logger.Info("TestDatabase: Initial drop of 'PhoneList' returned "+tResult.ToString());
                }
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: DropTable test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            // Test create new table, will fail if table cannot be created
            myTrans.Commit(); myTrans = null;
            logger.Info("TestDatabase:4: Testing CreateTable");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = PhoneList.CreateTable("PhoneList", myConnection, myTrans);
                if (IdLibrary.AnyIdError(tResult))
                {
                    logger.Info("TestDatabase: Create of 'PhoneList' failed (Cnt="+tResult.ToString()+")");
                    if (myTrans != null)
                    {
                        myTrans.Rollback(); myTrans = null;
                    }
                    myConnection.Close();
                    return false;
                }
                else
                {
                    logger.Info("TestDatabase: Create of 'PhoneList' success");
                }
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: CreateTable test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            // Test inserting record into table, will fail if insert fails
            myTrans.Commit(); myTrans = null;
            logger.Info("TestDatabase:5: Testing InsertRecord");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = sampleList.InsertRecord("PhoneList", myConnection, myTrans);
                if (IdLibrary.AnyIdError(tResult))
                {
                    logger.Info("TestDatabase: Insert of " + sampleList.GetCount() + " recs failed (Rslt="+tResult.ToString()+")");
                    if (myTrans != null)
                    {
                        myTrans.Rollback(); myTrans = null;
                    }
                    myConnection.Close();
                    return false;
                }

                myTrans.Commit(); myTrans = null;
                readList = new PhoneList();
                whereClause = "UpperID = '"+upperID+"'";
                readResult = readList.SqlSelectByWhere("PhoneList", myConnection, whereClause);

                if (readResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabase:5.4: Failed to get "+sampleList.GetCount()+" records from db, rslt="+readResult.ToString()+".");
                    myConnection.Close();
                    return false;
                }

                readList.Logit(LogLevel.Info, "", "TestDatabase: After insert", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: InsertRecord test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            // Test updating record into table, will fail if update fails
            logger.Info("TestDatabase:6: Testing UpdateRecordbyLowerId");
            try
            {
                if (sampleList.UpdateSample() == false)
                {
                    logger.Info("TestDatabase: Update Sample for 'PhoneList' failed ");
                    myConnection.Close();
                    return false;
                }

                myTrans = myConnection.BeginTransaction();
                tResult = sampleList.UpdateRecordByLowerId("PhoneList", myConnection, myTrans);
                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabase: Update of " + sampleList.GetCount() + " recs failed (Cnt="+tResult.ToString()+")");
                    if (myTrans != null)
                    {
                        myTrans.Rollback(); myTrans = null;
                    }
                    myConnection.Close();
                    return false;
                }

                sampleList.Logit(LogLevel.Info, "", "TestDatabase: After update sample:", IdLibrary.LogDocDepth.MAX);
                myTrans.Commit(); myTrans = null;

                readList = new PhoneList();
                whereClause = "UpperID = '"+upperID+"'";
                readResult = readList.SqlSelectByWhere("PhoneList", myConnection, whereClause);

                if (readResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabase:6.4: Failed to get "+sampleList.GetCount()+" records from db, readResult="+readResult.ToString()+".");
                    myConnection.Close();
                    return false;
                }

                readList.Logit(LogLevel.Info, "", "TestDatabase: After update:", IdLibrary.LogDocDepth.MAX);
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: UpdateRecord test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                myConnection.Close();
                return false;
            }

            if (aDeleteAtEnd == true)
            {
                if (TestDatabaseFinalRecordDelete(upperID, myConnection, sampleList, whereClause) == false)
                {
                    logger.Info("TestDatabase: TestDatabaseFinalTableDelete() failed");
                    myConnection.Close();
                    return false;
                }

                if (TestDatabaseDropTable(myConnection) == false)
                {
                    logger.Info("TestDatabase: TestDatabaseDropTable(PhoneItem) failed");
                    myConnection.Close();
                    return false;
                }
            }

            // Success
            logger.Info("TestDatabase: tests completed successfully");
            myConnection.Close();
            return true;
        }

        /// <summary>
        /// database final table delete test routine for PhoneList.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabaseFinalRecordDelete(string aUpperID, SqlConnection myConnection, PhoneList sampleList, string whereClause)
        {
            IdLibrary.TriResult tResult;
            PhoneList readList;
            SqlTransaction myTrans = null;

            // Test deleting record into table, will fail if insert fails
            logger.Info("TestDatabaseFinalRecordDelete:1: Testing DeleteRecordByLowerId");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = sampleList.DeleteRecordByLowerId("PhoneList", myConnection, myTrans);
                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabaseFinalRecordDelete: Delete of " + sampleList.GetCount() + " recs failed (Cnt="+tResult.ToString()+")");
                    if (myTrans != null)
                    {
                        myTrans.Rollback(); myTrans = null;
                    }
                    return false;
                }

                myTrans.Commit(); myTrans = null;
                readList = new PhoneList();
                whereClause = "UpperID = '"+aUpperID+"'";
                tResult = readList.SqlSelectByWhere("PhoneList", myConnection, whereClause);

                if (tResult != IdLibrary.TriResult.NOTFOUND)
                {
                    logger.Info("TestDatabaseFinalRecordDelete:4: Failed to delete "+sampleList.GetCount()+" records from db.");
                    return false;
                }

                return true;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabaseFinalRecordDelete: DeleteRecordByLowerId test got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                return false;
            }
        }

        /// <summary>
        /// database drop test routine for PhoneList.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabaseDropTable(SqlConnection myConnection)
        {
            IdLibrary.TriResult tResult;
            SqlTransaction myTrans = null;

            // Test drop existing table, will fail if table not present
            logger.Info("TestDatabaseDrop:1: Testing final DropTable");
            try
            {
                myTrans = myConnection.BeginTransaction();
                tResult = PhoneList.DropTable("PhoneList", myConnection, myTrans);
                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Info("TestDatabaseDrop: Final drop of 'PhoneList' returned "+tResult.ToString()+", returning false");
                    if (myTrans != null)
                    {
                        myTrans.Rollback(); myTrans = null;
                    }
                    return false;
                }

                // Success
                logger.Info("TestDatabaseDrop: test completed successfully");
                myTrans.Commit(); myTrans = null;
                return true;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabaseDrop: got ex: " + ex);
                if (myTrans != null)
                {
                    myTrans.Rollback(); myTrans = null;
                }
                return false;
            }
        }

        #endregion

        #region Class-Specific


        #endregion


        #endregion

        #region Private Access

        #region Private Members

        private string mUpperID;
        private string mLowerID;
        private List<PhoneItem> mTheList;
        private int mCurIdx;

        #endregion

        #endregion
    }
}
