using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using NLog;
using System.IO;
using System.Collections;
using System.Data.SqlClient;

namespace MyNameSpace
{
    [Serializable]
    public class FloorList
    {
        #region Constants

        public static string CLASS_NAME = typeof(FloorList).ToString();

        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        #endregion

        #region Public Access

        #region Public Properties

        public string UpperIDValue
        {
            get { return mUpperID; }
            set { mUpperID = value; }
        }

        public string LowerIDValue
        {
            get { return mLowerID; }
            set { mLowerID = value; }
        }

        public List<FloorItem> TheListValue
        {
            get { return mTheList; }
            set { mTheList = value; }
        }

        public int CurIdxValue
        {
            get { return mCurIdx; }
            set { mCurIdx = value; }
        }

        public string HeaderValue
        {
            get { return mHeader; }
            set { mHeader = value; }
        }

        [XmlIgnore]
        // Indexer to give array access to the main list.
        public FloorItem this[int index]
        {
            // Return the item at this index
            get
            {
                if (index < 0 || index >= mTheList.Count)
                    return null;

                return mTheList[index];
            }
            // Set the item at this index to given value
            set
            {
                if (index < 0 || index >= mTheList.Count)
                    return;

                mTheList[index] = value;
            }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Set default values for all fields.
        /// </summary>
        public FloorList()
        {
            Reset();
        }

        /// <summary>
        /// Set all fields according to arguments
        /// </summary>
        /// <param name=""></param>
        public FloorList(string aUpperID, string aLowerID, List<FloorItem> aTheList, int aCurIdx, string aHeader)
        {
            Set(aUpperID, aLowerID, aTheList, aCurIdx, aHeader);
        }

        /// <summary>
        /// Copy Constructor
        /// </summary>
        /// <param name=""></param>
        public FloorList(FloorList aItem)
        {
            Set(aItem);
        }

        #endregion

        #region Reset, Set

        /// <summary>
        /// Set fields to default values
        /// </summary>
        public void Reset()
        {
            mUpperID = "00000000-0000-0000-0000-000000000000";
            mLowerID = "00000000-0000-0000-0000-000000000000";
            mTheList = new List<FloorItem>();
            mCurIdx = -1;
            mHeader = "unused";
        }

        /// <summary>
        /// Set the values of every field according to arguments
        /// </summary>
        /// <param name=""></param>
        public void Set(string aUpperID, string aLowerID, List<FloorItem> aTheList, int aCurIdx, string aHeader)
        {
            mUpperID = aUpperID;
            mLowerID = aLowerID;
            mTheList = new List<FloorItem>(aTheList);
            mCurIdx = aCurIdx;
            mHeader = aHeader;
        }

        /// <summary>
        /// Copy our values from given FloorList
        /// </summary>
        /// <param name="aFrom"></param>
        public void Set(FloorList aFrom)
        {
            mUpperID = aFrom.mUpperID;
            mLowerID = aFrom.mLowerID;
            mTheList = new List<FloorItem>(aFrom.mTheList);
            mCurIdx = aFrom.mCurIdx;
            mHeader = aFrom.mHeader;
        }

        #endregion

        #region Standard Methods

        /// <summary>
        /// Add the given item to the end of the list.
        /// </summary>
        /// <param name="item"></param>
        public void AddItem(FloorItem item)
        {
            FloorItem tempItem = new FloorItem(item);
            tempItem.UpperIDValue = mLowerID;
            mTheList.Add(tempItem);
        }

        /// <summary>
        /// Return the number of items in the list.
        /// </summary>
        /// <returns></returns>
        public int GetCount()
        {
            return mTheList.Count;
        }

        /// <summary>
        /// Return the item at the requested index.
        /// </summary>
        /// <param name="idx"></param>
        /// <returns></returns>
        public FloorItem GetItem(int idx)
        {
            if (idx >= mTheList.Count)
                return null;

            return mTheList[idx];
        }

        /// <summary>
        /// Remove the item at the given index
        /// </summary>
        /// <param name="aIdx"></param>
        /// <returns></returns>
        public bool RemoveAt(int aIdx)
        {
            if (aIdx < 0 || aIdx >= mTheList.Count)
                return false;

            mTheList.RemoveAt(aIdx);
            return true;
        }

        /// <summary>
        /// Allows foreach to work with our list
        /// </summary>
        /// <returns></returns>
        public IEnumerator GetEnumerator()
        {
            foreach (object item in mTheList)
            {
                yield return item;
            }
        }

        #endregion

        #region Find / Get Methods

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aUpperID"></param>
        /// <returns></returns>
        public int FindUpperID(string aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                FloorItem item = mTheList[i];

                if (item.UpperIDValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aUpperID"></param>
        /// <returns></returns>
        public FloorItem GetUpperID(string aValue)
        {
            foreach (FloorItem item in mTheList)
            {
                if (item.UpperIDValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aLowerID"></param>
        /// <returns></returns>
        public int FindLowerID(string aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                FloorItem item = mTheList[i];

                if (item.LowerIDValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aLowerID"></param>
        /// <returns></returns>
        public FloorItem GetLowerID(string aValue)
        {
            foreach (FloorItem item in mTheList)
            {
                if (item.LowerIDValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Find idx of the desired Field type in the list. Return -1 if not found.
        /// </summary>
        /// <param name="aOffices"></param>
        /// <returns></returns>
        public int FindOffices(OfficeList aValue)
        {
            int i;

            for (i = 0; i < mTheList.Count; i++)
            {
                FloorItem item = mTheList[i];

                if (item.OfficesValue == aValue)
                    return i;
            }

            // Not found
            return -1;
        }

        /// <summary>
        /// Return first item in the list with this value.
        /// </summary>
        /// <param name="aOffices"></param>
        /// <returns></returns>
        public FloorItem GetOffices(OfficeList aValue)
        {
            foreach (FloorItem item in mTheList)
            {
                if (item.OfficesValue == aValue)
                    return item;
            }

            // Not found
            return null;
        }

        /// <summary>
        /// Return item at the current index, or null if none
        /// </summary>
        /// <returns></returns>
        public FloorItem GetCurrentItem()
        {
            if (mCurIdx < 0 || mCurIdx >= mTheList.Count)
                return null;

            return mTheList[mCurIdx];
        }

        #endregion

        #region Logging

        /// <summary>
        /// Return ToString for this class, or "null"
        /// </summary>
        /// <param name="aValue"></param>
        /// <returns></returns>
        public static string ToStringP(FloorList aValue)
        {
            if (aValue == null)
                return "null";
            else
                return aValue.ToString();
        }

        /// <summary>
        /// Return fields of this object as a string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "mUpperID="+mUpperID+", mLowerID="+mLowerID+", mTheList='"+mTheList.ToString()+"'"+", mCurIdx="+mCurIdx+", mHeader='"+mHeader+"'";
        }

        /// <summary>
        /// Log this object to nlog
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void Logit(LogLevel aLevel, string aIndent, string aHeader)
        {
            string logString = LogitToString(aLevel, aIndent, aHeader);
            logger.Log(aLevel, logString);
        }

        /// <summary>
        /// Log this object to nlog - short version
        /// </summary>
        /// <param name="aLevel"></param>
        /// <param name="aIndent"></param>
        /// <param name="aHeader"></param>
        public void LogitL(LogLevel aLevel, string aIndent, string aHeader)
        {
            string logString = LogitLToString(aLevel, aIndent, aHeader);
            logger.Log(aLevel, logString);
        }

        /// <summary>
        /// Log the contents of this class to String
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitToString(LogLevel aLevel, string aIndent, string aHeader)
        {
            // Don't log if not within level
            if (WithinLoggerLevel(aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            sb.AppendLine();
            sb.AppendLine(aIndent+aHeader + " FloorList. CurIdx=" + mCurIdx + ", item count=" + mTheList.Count + ":" );

            int i;

            for (i = 0; i < mTheList.Count; i++)
                sb.AppendLine(mTheList[i].LogitToString(aLevel, aIndent + "    ", "[" + i + "]:"));

            sb.AppendLine(aIndent+"    "+"mUpperID="+mUpperID+", mLowerID="+mLowerID+", mHeader='"+mHeader+"'");

            sb.Append(aIndent + "End of FloorList");
            return sb.ToString();
        }

        /// <summary>
        /// Log the contents of this class to String - short version
        /// </summary>
        /// <param name="aHeader">Header info: EventItem: or [0]: or whatever</param>
        public string LogitLToString(LogLevel aLevel, string aIndent, string aHeader)
        {
            // Don't log if not within level
            if (WithinLoggerLevel(aLevel) == false)
                return "";

            StringBuilder sb = new StringBuilder();

            sb.AppendLine();
            sb.AppendLine(aIndent+aHeader + " FloorList. CurIdx=" + mCurIdx + ", item count=" + mTheList.Count + ":" );

            int i;

            for (i = 0; i < mTheList.Count; i++)
                sb.AppendLine(mTheList[i].LogitLToString(aLevel, aIndent + "    ", "[" + i + "]:"));

            sb.AppendLine(aIndent+"    "+"mUpperID="+mUpperID+", mLowerID="+mLowerID+", mHeader='"+mHeader+"'");

            sb.Append(aIndent + "End of FloorList");
            return sb.ToString();
        }

        #endregion

        #region I/O Routines

        /// <summary>
        /// Convert our class to an XML string via Serialize. Can throw exceptions
        /// </summary>
        /// <returns></returns>
        public string ToXML()
        {
            try
            {
                // Force UTF-8
                XmlSerializer serializer = new XmlSerializer(typeof(FloorList));
                using (StringWriter writer = new Utf8StringWriter())
                {
                    serializer.Serialize(writer, this);
                    return writer.ToString();
                }
            }
            catch (Exception ex)
            {
                logger.Error("ToXML: Exception '" + ex + "'");
                throw new Exception("ToXML", ex);
            }
        }

        /// <summary>
        /// Convert given xml string to our class. Throw exception on fail.
        /// </summary>
        /// <param name="xml"></param>
        /// <returns></returns>
        public FloorList FromXML(string xml)
        {
            FloorList ei = null;
            if ((xml != null) && (xml.Length > 0))
            {
                try
                {
                    XmlSerializer serializer = new XmlSerializer(typeof(FloorList));
                    ei = (FloorList)serializer.Deserialize(new StringReader(xml));
                }
                catch (Exception ex)
                {
                    logger.Error("FromXML: Got exception '" + ex + "'");
                    throw new Exception("FromXML", ex);
                }
            }
            return ei;
        }

        /// <summary>
        /// Write out our class to given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool WriteFile(string fileName)
        {
            try
            {
                StreamWriter writer = new StreamWriter(fileName);
                writer.Write(this.ToXML());
                writer.Close();
            }
            catch (Exception ex)
            {
                logger.Info("WriteFile: Got exception '" + ex + "'");
                return false;
            }
            // Success
            return true;
        }

        /// <summary>
        /// Read our class from given file. Return true on success, false on fail.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public bool ReadFile(string fileName)
        {
            try
            {
                StreamReader reader = new StreamReader(fileName);
                string contents = reader.ReadToEnd();
                reader.Close();

                // Set all our values from what we got back
                FloorList fi = FromXML(contents);
                this.Set(fi);
            }
            catch (Exception ex)
            {
                logger.Info("ReadFile: Got exception '" + ex + "'");
                return false;
            }

            // Success
            return true;
        }

        #endregion

        #region Common Routines

        // Force utf-8 instead of 16
        public class Utf8StringWriter : StringWriter
        {
            public override Encoding Encoding
            {
                get { return Encoding.UTF8; }
            }
        }

        #endregion

        #region Tester

        /// <summary>
        /// Basic test routine item and list.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool Tester()
        {
            string TEST_FILE = "testFloor.xml";
            FloorItem item1, item2, item3;
            FloorList list1, list2;
            int findIdx;
            bool result;
            string upperID;
            string listLowerID;

            // Test creating the items and building a list
            try
            {
                logger.Info("");

                // Create list #1
                upperID = Guid.NewGuid().ToString();
                list1 = new FloorList("e5b4218d-bf0d-4650-bdca-ca105dc968d2", "94d212c4-f022-41fa-a448-9e7c96935c07", new List<FloorItem>(), 2, "bbb");
                listLowerID = list1.LowerIDValue;
                item1 = FloorItem.BuildSample(listLowerID, 1);
                list1.AddItem(item1);
                logger.Info("Tester:2: 1 item list string=[" + list1.ToString() + "]");

                // Create list #2
                upperID = Guid.NewGuid().ToString();
                list2 = new FloorList(list1);
                list2.LowerIDValue = Guid.NewGuid().ToString();
                listLowerID = list2.LowerIDValue;

                foreach (FloorItem updateItem in list2)
                    updateItem.UpperIDValue = listLowerID;

                item2 = FloorItem.BuildSample(listLowerID, 2);
                list2.AddItem(item2);

                item3 = FloorItem.BuildSample(listLowerID, 3);
                list2.AddItem(item3);

                list2.Logit(LogLevel.Info, "", "Tester:3: List2:");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Build list section got ex: " + ex);
                return false;
            }

            // Test logging and finds
            try
            {
                list2.Logit(LogLevel.Info, "", "Tester:4: created list:");

                // Supposed to fail
                findIdx = list2.FindUpperID("00000000-0000-0000-0000-000000000000");

                if (findIdx >= 0)
                {
                    logger.Info("Tester: FindUpperID(00000000-0000-0000-0000-000000000000) succeeded, returning false");
                    return false;
                }

                // Supposed to succeed
                findIdx = list2.FindUpperID(list2[0].UpperIDValue);

                if (findIdx < 0)
                {
                    logger.Info("Tester:5: FindUpperID(list2[0].UpperIDValue) failed, returning false");
                    return false;
                }

                // Test removal
                result = list2.RemoveAt(findIdx);

                if (result != true)
                {
                    logger.Info("Tester:6: list2.RemoveAt(" + findIdx + ") failed, returning false");
                    return false;
                }

                list2.LogitL(LogLevel.Info, "", "Tester: After remove, list2:");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Build list section got ex: " + ex);
                return false;
            }

            try
            {
                // Write list2 out to XML
                if (list2.WriteFile(TEST_FILE) == false)
                {
                    logger.Info("Tester:7: list2 WriteFile("+TEST_FILE+") failed, returning false");
                    return false;
                }

                // Read in as ul1
                result = list1.ReadFile(TEST_FILE);

                if (result != true)
                {
                    logger.Info("Tester:8: ReadFile(" + TEST_FILE + ") failed, returning false");
                    return false;
                }

                list1.Logit(LogLevel.Info, "", "Tester:9: Read "+TEST_FILE+":");
            }
            catch (Exception ex)
            {
                logger.Error("Tester: Read/Write list section got ex: " + ex);
                return false;
            }

            // Success
            return true;
        }

        /// <summary>
        /// Return a sample record FloorList.
        /// Return sample on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static FloorList BuildSample(string aUpperID, int aValue)
        {
            FloorItem item;
            FloorList rec;

            rec = new FloorList("70dc998d-5382-452e-a92c-c73ef41a5eb7", "3d189e64-8258-4c41-9d78-5a9d50e5ba97", new List<FloorItem>(), 1, "aaa");
            rec.UpperIDValue = aUpperID;
            rec.LowerIDValue = Guid.NewGuid().ToString();

            // Build list of 3 of same kind
            if (aValue == 1 || aValue == 2 || aValue == 3)
            {
                item = FloorItem.BuildSample(rec.LowerIDValue, aValue);
                rec.AddItem(item);
                item = FloorItem.BuildSample(rec.LowerIDValue, aValue);
                rec.AddItem(item);
                item = FloorItem.BuildSample(rec.LowerIDValue, aValue);
                rec.AddItem(item);
            }
            else
            {
                item = FloorItem.BuildSample(rec.LowerIDValue, -1);
                rec.AddItem(item);
                item = FloorItem.BuildSample(rec.LowerIDValue, -1);
                rec.AddItem(item);
                item = FloorItem.BuildSample(rec.LowerIDValue, -1);
                rec.AddItem(item);
            }

            return rec;
        }

        /// <summary>
        /// Drill down through sample record until non-class 2nd item and update it
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public bool UpdateSample()
        {
            logger.Trace("UpdateSample:1:");

            foreach (FloorItem rec in mTheList)
            {
                if (rec.UpdateSample() == false)
                    return false;
            }

            return true;
        }

        #endregion

        #region Database Routines

        #region Table Level Routines

        /// <summary>
        /// Create given table in database.
        /// Return 1 on success (1 rec), -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public static int CreateTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            int recCount = 0, totRecCount=0;

            try
            {
                string cmdHeader = "CREATE TABLE "+aTableName+"(";
                string cmdArgs = "UpperID UNIQUEIDENTIFIER DEFAULT NEWID(), LowerID UNIQUEIDENTIFIER DEFAULT NEWID(), CurIdx int, Header varchar(255)";
                string cmdClose = ")";
                string command = cmdHeader + cmdArgs + cmdClose;
                recCount = RunSqlNonQuery(command, aConnection, aTrans);
                logger.Trace("CreateTable: RunSql returned "+recCount);

                // Unfortunately, create table returns -1 on success or fail. Check for exception (-2)
                if (recCount == -2)
                    return -1;

                // Add each class as a call for another table
                recCount = FloorItem.CreateTable("FloorItem", aConnection, aTrans);
                if (recCount < 0)
                    return -1;

                return 1;
            }
            catch (Exception ex)
            {
                logger.Error("CreateTable: Got ex: "+ex);
                return -1;
            }
        }

        /// <summary>
        /// Destroy given table in database.
        /// Return 1 on success (1 rec), -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to drop</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public static int DropTable(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            int recCount = 0, totRecCount=0;

            try
            {
                string command = "DROP TABLE dbo."+aTableName;
                recCount = RunSqlNonQuery(command, aConnection, aTrans);
                logger.Trace("DropTable: RunSql returned "+recCount);

                // Unfortunately, drop table returns -1 on success or fail. Check for exception (-2)
                if (recCount == -2)
                    return -1;

                // Add each class as a call for another table
                recCount = FloorItem.DropTable("FloorItem", aConnection, aTrans);

                if (recCount < 0)
                    return -1;

                return 1;
            }
            catch (Exception ex)
            {
                logger.Error("DropTable: Got ex: "+ex);
                return -1;
            }
        }

        #endregion

        #region DB Record Insert/Update/Delete Routines

        /// <summary>
        /// Insert list records into database.
        /// Return 1 on success (1 rec), -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public int InsertRecord(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            int recCount = 0, totRecCount = 0;

            try
            {
                string cmdHeader = "INSERT INTO "+aTableName+" (";
                string columnArgs = "UpperID, LowerID, CurIdx, Header";
                string colClose = ") Values (";
                string valueArgs = "'"+mUpperID+"', '"+mLowerID+"', "+mCurIdx+", '"+mHeader+"'";
                string cmdClose = ");";
                string command = cmdHeader + columnArgs + colClose + valueArgs + cmdClose;

                recCount = RunSqlNonQuery(command, aConnection, aTrans);
                logger.Info("InsertRecord:2: RunSqlNonQuery returned "+recCount);
                totRecCount += recCount;

                // Add each class as a call for another table
                foreach (FloorItem item in mTheList)
                {
                    recCount = item.InsertRecord("FloorItem", aConnection, aTrans);

                    if (recCount < 1)
                    {
                        item.Logit(LogLevel.Error, "", "item.InsertRecord: Failed to insert:");
                        return -1;
                    }

                    totRecCount += recCount;
                }

                return totRecCount;
            }
            catch (Exception ex)
            {
                logger.Error("InsertRecord: Got ex: "+ex);
                return -1;
            }
        }

        /// <summary>
        /// Update all fields of all list records in database.
        /// Return count updated on success , -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public int UpdateRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            int recCount = 0, totRecCount=0;

            try
            {
                string cmdHeader = "UPDATE "+aTableName+" SET ";
                string columnArgs = "UpperID = '"+mUpperID+"', LowerID = '"+mLowerID+"', CurIdx = "+mCurIdx+", Header = '"+mHeader+"'";
                string colClose = " WHERE LowerID =";
                string valueArgs = "'"+mLowerID+"'";
                string cmdClose = ";";
                string command = cmdHeader + columnArgs + colClose + valueArgs + cmdClose;

                recCount = RunSqlNonQuery(command, aConnection, aTrans);
                logger.Info("UpdateRecord:2: RunSqlNonQuery returned "+recCount);
                totRecCount += recCount;

                // Add each class as a call for another table
                foreach (FloorItem item in mTheList)
                {
                    recCount = item.UpdateRecordByLowerId("FloorItem", aConnection, aTrans);

                    if (recCount < 1)
                    {
                        item.Logit(LogLevel.Error, "", "item.UpdateRecordByLowerId: Failed to insert:");
                        return -1;
                    }

                    totRecCount += recCount;
                }

                return totRecCount;
            }
            catch (Exception ex)
            {
                logger.Error("UpdateRecordByLowerId: Got ex: "+ex);
                return -1;
            }
        }

        /// <summary>
        /// Delete this record from database.
        /// Return count deleted on success , -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public int DeleteRecordByWhere(string aTableName, SqlConnection aConnection, SqlTransaction aTrans, string aWhereClause)
        {
            int recCount = 0, totRecCount=0;

            try
            {
                // Add each class as a call for another table
                foreach (FloorItem item in mTheList)
                {
                    recCount = item.DeleteRecordByLowerId("FloorItem", aConnection, aTrans);

                    if (recCount < 1)
                    {
                        item.Logit(LogLevel.Error, "", "item.DeleteRecordByLowerId: Failed to delete:");
                        return -1;
                    }

                    totRecCount += recCount;
                }
                string cmdHeader = "DELETE FROM "+aTableName+" WHERE ";
                string valueArgs = aWhereClause;
                string cmdClose = ";";
                string command = cmdHeader + valueArgs + cmdClose;

                recCount = RunSqlNonQuery(command, aConnection, aTrans);
                if (recCount < 0)
                {
                    logger.Error("DeleteRecordByWhere: RunSqlNonQuery failed");
                    return -1;
                }

                logger.Trace("DeleteRecordByWhere: RunSql returned "+recCount);
                totRecCount += recCount;
                return totRecCount;
            }
            catch (Exception ex)
            {
                logger.Error("DeleteRecordByWhere: Got ex: "+ex);
                return -1;
            }
        }

        /// <summary>
        /// Delete this record from database.
        /// Return count deleted on success , -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public int DeleteRecordByLowerId(string aTableName, SqlConnection aConnection, SqlTransaction aTrans)
        {
            return DeleteRecordByWhere(aTableName, aConnection, aTrans, "LowerID = '"+ mLowerID + "'");
        }

        #endregion

        #region DB Record Select Open/Read Routines

        /// <summary>
        /// Create and open the SQL select to get fields of this record.
        /// Return reader on success, null on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public SqlDataReader SqlOpen(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            SqlDataReader myReader = null;

            try
            {
                string cmdHeader = "SELECT ";
                string columnArgs = "UpperID, LowerID, CurIdx, Header";
                string colClose = " FROM "+aTableName+" WHERE " + aWhereClause + ";";
                string command = cmdHeader + columnArgs + colClose;

                logger.Trace("SqlOpen: command=["+command+"]");
                SqlCommand myCommand = new SqlCommand(command, aConnection);
                myReader = myCommand.ExecuteReader();
                return myReader;
            }
            catch (Exception ex)
            {
                logger.Error("SqlOpen: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return null;
            }
        }

        /// <summary>
        /// Read in the next sql record into this Item.
        /// Return true on success, false on fail
        /// </summary>
        /// <param name="aReader">Reader to read from</param>
        public bool SqlRead(SqlDataReader myReader)
        {
            bool result;

            try
            {
                result = myReader.Read();
                if (result == false)
                    return false;

                mUpperID = myReader.GetGuid(0).ToString();
                mLowerID = myReader.GetGuid(1).ToString();
                mTheList = new List<FloorItem>();
                mCurIdx = myReader.GetInt32(2);
                mHeader = myReader.GetString(3);

                return true;
            }
            catch (Exception ex)
            {
                logger.Error("SqlRead: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return false;
            }
        }

        #endregion

        #region DB Record Select Routines

        /// <summary>
        /// Select multiple records from database.
        /// Return num of rec (list rec+direct items) on success, -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public int SqlSelectByLowerId(string aTableName, SqlConnection aConnection)
        {
            return SqlSelectByWhere(aTableName, aConnection, "LowerID = '" + mLowerID+"'");
        }

        /// <summary>
        /// Select multiple records from database.
        /// Return num of rec (list rec + direct items) on success, -1 on fail
        /// </summary>
        /// <param name="aTableName">Table name to create</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public int SqlSelectByWhere(string aTableName, SqlConnection aConnection, string aWhereClause)
        {
            SqlDataReader myReader = null;
            bool result;

            try
            {
                logger.Trace("SqlSelectByWhere:1: aTableName='"+aTableName+"', aWhereClause='"+aWhereClause+"'");

                // Create and open the appropriate select
                myReader = SqlOpen(aTableName, aConnection, aWhereClause);
                if (myReader == null)
                    return -1;

                // Since this is our fields, get exactly 1 record
                result = SqlRead(myReader);
                logger.Trace("SqlSelectByWhere:2: SqlRead returned "+result);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                if (result == false)
                    return -1;

                return SqlSelectSubrecords(aConnection)+1;
            }
            catch (Exception ex)
            {
                logger.Error("SqlSelectByWhere: Got ex: "+ex);

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                return -1;
            }
        }

        /// <summary>
        /// Load subclasses with their sql data.
        /// Return return count of direct subrecords on success, -1 on fail
        /// </summary>
        public int SqlSelectSubrecords(SqlConnection aConnection)
        {
            SqlDataReader myReader = null;
            int recCount = 0;

            try
            {
                // Add each class as a call for another table
                string whereClause = "UpperID = '"+mLowerID+"'";
                logger.Trace("SqlSelectSubrecords:1: whereClause='"+whereClause+"'");

                FloorItem listItem = new FloorItem();

                // Create and open the appropriate select
                myReader = listItem.SqlOpen("FloorItem", aConnection, whereClause);
                if (myReader == null)
                {
                    logger.Info("SqlSelectSubrecords: SqlOpen(FloorItem) failed");
                    return -1;
                }

                // Read until run out of records
                while (listItem.SqlRead(myReader))
                {
                    mTheList.Add(new FloorItem(listItem));
                    recCount++;
                }

                if (myReader != null && myReader.IsClosed == false)
                    myReader.Close();

                foreach (FloorItem item in mTheList)
                    if (item.SqlSelectSubrecords(aConnection) < 0)
                        return -1;

                return recCount;
            }
            catch (Exception ex)
            {
                logger.Error("SqlSelectSubrecords: Got ex: "+ex);
                return -1;
            }
        }

        #endregion

        #endregion

        #region TestDatabase

        /// <summary>
        /// Basic database test routine for FloorList.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabase(bool aDeleteAtEnd)
        {
            int recCount, readCount;
            SqlConnection myConnection;
            SqlTransaction myTrans = null;
            string dbName = "OfficeDemo";
            FloorList sampleList, readList;
            string whereClause;
            string upperID = Guid.NewGuid().ToString();

            logger.Info("");
            logger.Info("TestDatabase:1: upperID="+upperID);

            // Test building sample list
            try
            {
                sampleList = BuildSample(upperID, 1);
                sampleList.Logit(LogLevel.Info, "", "TestDatabase: Sample list:");
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: Build sample list test got ex: " + ex);
                return false;
            }

            // Test database open
            logger.Info("TestDatabase:2: Testing OpenDb");
            try
            {
                myConnection = OpenDbConnection(null, dbName);
                if (myConnection == null)
                {
                    logger.Error("TestDatabase: Can't open connection to database '"+dbName+"'");
                    return false;
                }
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: Open test got ex: " + ex);
                return false;
            }

            // Test drop existing tables, ignore tables not present
            logger.Info("TestDatabase:3: Testing initial DropTable");
            try
            {
                recCount = FloorList.DropTable("FloorList", myConnection, myTrans);
                if (recCount < 1)
                {
                    logger.Info("TestDatabase: Initial drop of 'FloorList' returned "+recCount+", ignore");
                }
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: DropTable test got ex: " + ex);
                myConnection.Close();
                return false;
            }

            // Test create new table, will fail if table cannot be created
            logger.Info("TestDatabase:4: Testing CreateTable");
            try
            {
                recCount = FloorList.CreateTable("FloorList", myConnection, myTrans);
                if (recCount < 1)
                {
                    logger.Info("TestDatabase: Create of 'FloorList' failed (Cnt="+recCount+")");
                    myConnection.Close();
                    return false;
                }
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: CreateTable test got ex: " + ex);
                myConnection.Close();
                return false;
            }

            // Test inserting record into table, will fail if insert fails
            logger.Info("TestDatabase:5: Testing InsertRecord");
            try
            {
                recCount = sampleList.InsertRecord("FloorList", myConnection, myTrans);
                if (recCount < sampleList.GetCount()+1)
                {
                    logger.Info("TestDatabase: Insert of " + sampleList.GetCount() + " recs failed (Cnt="+recCount+")");
                    myConnection.Close();
                    return false;
                }

                readList = new FloorList();
                whereClause = "UpperID = '"+upperID+"'";
                readCount = readList.SqlSelectByWhere("FloorList", myConnection, whereClause);

                if (readCount != sampleList.GetCount()+1)
                {
                    logger.Info("TestDatabase:5.4: Failed to get "+sampleList.GetCount()+" records from db, readCount="+readCount+".");
                    myConnection.Close();
                    return false;
                }

                readList.Logit(LogLevel.Info, "", "TestDatabase: After insert");
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: InsertRecord test got ex: " + ex);
                myConnection.Close();
                return false;
            }

            // Test updating record into table, will fail if update fails
            logger.Info("TestDatabase:6: Testing UpdateRecordbyLowerId");
            try
            {
                myTrans = myConnection.BeginTransaction();
                if (sampleList.UpdateSample() == false)
                {
                    logger.Info("TestDatabase: Update Sample for 'FloorList' failed ");
                    myConnection.Close();
                    return false;
                }

                recCount = sampleList.UpdateRecordByLowerId("FloorList", myConnection);
                if (recCount < sampleList.GetCount())
                {
                    logger.Info("TestDatabase: Update of " + sampleList.GetCount() + " recs failed (Cnt="+recCount+")");
                    myTrans.Rollback();
                    myConnection.Close();
                    return false;
                }

                sampleList.Logit(LogLevel.Info, "", "TestDatabase: After update sample:");

                readList = new FloorList();
                whereClause = "UpperID = '"+upperID+"'";
                readCount = readList.SqlSelectByWhere("FloorList", myConnection, whereClause);

                if (readCount != sampleList.GetCount()+1)
                {
                    logger.Info("TestDatabase:6.4: Failed to get "+sampleList.GetCount()+" records from db, readCount="+readCount+".");
                    myTrans.Rollback();
                    myConnection.Close();
                    return false;
                }

                readList.Logit(LogLevel.Info, "", "TestDatabase: After update:");
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabase: UpdateRecord test got ex: " + ex);
                myTrans.Rollback();
                myConnection.Close();
                return false;
            }

            if (aDeleteAtEnd == true)
            {
                if (TestDatabaseFinalRecordDelete(upperID, myConnection, sampleList, whereClause) == false)
                {
                    logger.Info("TestDatabase: TestDatabaseFinalTableDelete() failed");
                    myConnection.Close();
                    return false;
                }

                if (TestDatabaseDropTable(myConnection) == false)
                {
                    logger.Info("TestDatabase: TestDatabaseDropTable(FloorItem) failed");
                    myConnection.Close();
                    return false;
                }
            }

            // Success
            logger.Info("TestDatabase: tests completed successfully");
            myConnection.Close();
            return true;
        }

        /// <summary>
        /// database final table delete test routine for FloorList.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabaseFinalRecordDelete(string aUpperID, SqlConnection myConnection, FloorList sampleList, string whereClause)
        {
            int recCount = 0;
            FloorList readList;
            SqlTransaction myTrans = null;

            // Test deleting record into table, will fail if insert fails
            logger.Info("TestDatabaseFinalRecordDelete:1: Testing DeleteRecordByLowerId");
            try
            {
                myTrans = myConnection.BeginTransaction();
                recCount = sampleList.DeleteRecordByLowerId("FloorList", myConnection, myTrans);
                if (recCount < sampleList.GetCount())
                {
                    logger.Info("TestDatabaseFinalRecordDelete: Delete of " + sampleList.GetCount() + " recs failed (Cnt="+recCount+")");
                    myTrans.Rollback();
                    return false;
                }

                readList = new FloorList();
                whereClause = "UpperID = '"+aUpperID+"'";
                recCount = readList.SqlSelectByWhere("FloorList", myConnection, whereClause);

                if (recCount > 0)
                {
                    logger.Info("TestDatabaseFinalRecordDelete:4: Failed to delete "+sampleList.GetCount()+" records from db.");
                    myTrans.Rollback();
                    return false;
                }

                myTrans.Commit();
                return true;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabaseFinalRecordDelete: DeleteRecordByLowerId test got ex: " + ex);
                myTrans.Rollback();
                return false;
            }
        }

        /// <summary>
        /// database drop test routine for FloorList.
        /// Return true on success, false on fail.
        /// </summary>
        /// <returns></returns>
        public static bool TestDatabaseDropTable(SqlConnection myConnection)
        {
            int recCount = 0;
            SqlTransaction myTrans = null;

            // Test drop existing table, will fail if table not present
            logger.Info("TestDatabaseDrop:1: Testing final DropTable");
            try
            {
                myTrans = myConnection.BeginTransaction();
                recCount = FloorList.DropTable("FloorList", myConnection, myTrans);
                if (recCount < 1)
                {
                    logger.Info("TestDatabaseDrop: Final drop of 'FloorList' returned "+recCount+", returning false");
                    myTrans.Rollback();
                    return false;
                }

                // Success
                logger.Info("TestDatabaseDrop: test completed successfully");
                myTrans.Commit();
                return true;
            }
            catch (Exception ex)
            {
                logger.Error("TestDatabaseDrop: got ex: " + ex);
                myTrans.Rollback();
                return false;
            }
        }

        #endregion

        #region Database Support

        /// <summary>
        /// Return SqlConnection on success, null on fail.
        /// </summary>
        /// <param name="aSourceName">SQL server to use or null for default</param>
        /// <param name="aDbName">Database on server to connect to</param>
        /// <returns></returns>
        public static SqlConnection OpenDbConnection(string aSourceName, string aDbName)
        {
            try
            {
                logger.Info("OpenDbConnection:1: aSourceName='"+aSourceName+"', aDbName='"+aDbName+"'");

                SqlConnectionStringBuilder bldr = new SqlConnectionStringBuilder();
                SqlConnection myConnection;

                if (aSourceName == null)
                    bldr.DataSource = "localhost\\SQLExpress"; // Sql server to use
                else
                    bldr.DataSource = aSourceName+"\\SQLExpress"; // Sql server to use

                bldr.InitialCatalog = aDbName;     //The database on the server that you want to connect to.
                bldr.UserID = "OfficeDemoUser"; //The user id
                bldr.Password = "officeDemo";  //The pwd for said user account

                myConnection = new SqlConnection(bldr.ConnectionString);

                // Open the connection.
                myConnection.Open();
                return myConnection;
            }
            catch (Exception ex)
            {
                logger.Error("OpenDbConnection: Got ex: "+ex);
                return null;
            }
        }

        /// <summary>
        /// Run the given SQL command.
        /// Return rec count on success, -1 on fail, -2 on exception
        /// </summary>
        /// <param name="aCommand">Command to run</param>
        /// <param name="aConnection">Connection to use, or null for own open/close</param>
        public static int RunSqlNonQuery(string aCommand, SqlConnection aConnection, SqlTransaction aTrans)
        {
            SqlConnection myConnection = null;
            int recCount;
            SqlCommand myCommand;

            try
            {
                logger.Trace("RunSqlNonQuery:1: aCommand='"+aCommand+"'");

                if (aConnection != null)
                    myConnection = aConnection;
                else
                    myConnection = OpenDbConnection(null, "OfficeDemo");

                if (myConnection == null)
                    return -1;

                if (aTrans != null)
                    myCommand = new SqlCommand(aCommand, myConnection, aTrans);
                else
                    myCommand = new SqlCommand(aCommand, myConnection);

                recCount = myCommand.ExecuteNonQuery();

                if (aConnection == null)
                    myConnection.Close();

                logger.Info("RunSqlNonQuery:3: aCommand='"+aCommand+"', returned cnt="+recCount);
                return recCount;
            }
            catch (Exception ex)
            {
                // Don't want this common problem dumped in the error log
                if (ex.Message.Contains("because it does not exist") == false)
                    logger.Error("RunSqlNonQuery: Got ex: "+ex.Message+", returning -2");
                else
                    logger.Info("RunSqlNonQuery: Got ex: "+ex.Message+", returning -2");
                logger.Info("RunSqlNonQuery: aCommand was'"+aCommand+"'");

                if (aConnection == null && myConnection != null)
                    myConnection.Close();
                return -2;
            }
        }

        #endregion

        #region Class-Specific

        #endregion

        #endregion

        #region Private Access

        #region Private Members

        private string mUpperID;
        private string mLowerID;
        private List<FloorItem> mTheList;
        private int mCurIdx;
        private string mHeader;

        #endregion

        #region Support

        /// <summary>
        /// Return true if we should try to log, false if not.
        /// </summary>
        /// <param name="aLogger"></param>
        /// <param name="aLevel"></param>
        /// <returns></returns>
        private static bool WithinLoggerLevel(LogLevel aLevel)
        {
            if (aLevel == LogLevel.Trace && logger.IsTraceEnabled == true)
                return true;

            else if (aLevel == LogLevel.Debug && logger.IsDebugEnabled == true)
                return true;

            else if (aLevel == LogLevel.Info && logger.IsInfoEnabled == true)
                return true;

            else if (aLevel == LogLevel.Warn && logger.IsWarnEnabled == true)
                return true;

            else if (aLevel == LogLevel.Error)
                return true;

            else
                return false;
        }

        #endregion

        #endregion
    }
}
