﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NLog;
using IdcgLibrary;

// Uncomment only if database was checked
using System.Data.SqlClient;

namespace MyNameSpace
{
    class Program
    {
        // Effectively a constant
        private static Logger logger = LogManager.GetCurrentClassLogger();

        static void Main(string[] args)
        {
            //bool result;

            // Read in sample XML file, store in database (only uncomment if database was checked)
            logger.Info("Main:1: Reading XML file");
            InsertSampleFloor(false);

            //logger.Info("Main:2: Running FloorItem TestDatabase");
            //result = FloorItem.TestDatabase(false, "OfficeDemo", "OfficeDemoUser", "officeDemo");
            //logger.Info("Main:3: Finished Running FloorItem TestDatabase, result = " + result);

            // Sample routine to demonstrate calling a Class-Specific function that has been preserved
            // in the text files.
            logger.Info("Main:1: Calling BooWindows");
            BooWindows();

            logger.Info("Main:5: Finished Test\r\n\r\n");
        }


        /// <summary>
        /// Silly sample routine to demonstrate class specific function.
        /// </summary>
        private static void BooWindows()
        {
            bool result;

            // Read in sample XML file
            logger.Info("BooWindows:1: Reading XML file");
            FloorItem fi = new FloorItem();
            result = fi.ReadFile("sampleFloor.xml");

            if (result == false)
            {
                logger.Error("BooWindows: Failed to read in sampleFloor");
                return;
            }

            fi.Logit(LogLevel.Info, "", "After read:", IdLibrary.LogDocDepth.FULL);

            // Call Class-Specific function
            fi.OfficesValue[1].BooWindows();

            fi.Logit(LogLevel.Info, "", "After booWindows:", IdLibrary.LogDocDepth.MAX);

        }

        // Uncomment only if database version is used (checkbox)
        /// <summary>
        /// Test reading in sample XML and inserting it into database
        /// </summary>
        private static void InsertSampleFloor(bool aDeleteAtEnd)
        {
            bool result;
            IdLibrary.TriResult tResult;

            // Read in sample XML file
            logger.Info("InsertSampleFloor:1: Reading XML file");
            FloorItem fi = new FloorItem();
            result = fi.ReadFile("sampleFloor.xml");

            if (result == false)
            {
                logger.Error("InsertSampleFloor: Failed to read in sampleFloor");
                return;
            }

            fi.Logit(LogLevel.Info, "", "After read:", IdLibrary.LogDocDepth.MAX);

            // Open database
            SqlConnection myConnection;

            logger.Info("");
            myConnection = IdLibrary.OpenDbConnection(null, "OfficeDemo", "OfficeDemoUser", "officeDemo");
            if (myConnection == null)
            {
                logger.Error("InsertSampleFloor: Can't open connection to database 'OfficeDemo'");
                return;
            }

            // Begin transaction
            SqlTransaction myTrans = myConnection.BeginTransaction();

            // Drop existing floorItem and other tables
            tResult = FloorItem.DropTable("FloorItem", myConnection, myTrans);
            if (tResult != IdLibrary.TriResult.SUCCESS)
            {
                logger.Info("InsertSampleFloor: Can't drop FloorItem, ignoring");
            }

            // Commit on success
            myTrans.Commit();

            // Add new floorItem and another tables
            myTrans = myConnection.BeginTransaction();
            tResult = FloorItem.CreateTable("FloorItem", myConnection, myTrans);
            if (tResult != IdLibrary.TriResult.SUCCESS)
            {
                logger.Error("InsertSampleFloor: Failed to create FloorItem table");
                myTrans.Rollback();
                myConnection.Close();
                return;
            }

            // Commit on success
            myTrans.Commit();

            // Insert our record
            myTrans = myConnection.BeginTransaction();
            tResult = fi.InsertRecord("FloorItem", myConnection, myTrans);

            if (tResult != IdLibrary.TriResult.SUCCESS)
            {
                logger.Error("InsertSampleFloor: Failed to insert FloorItem record");
                myTrans.Rollback();
                myConnection.Close();
                return;
            }

            // Read in this series
            myTrans.Commit();

            FloorItem readItem = new FloorItem();
            readItem.LowerIDValue = fi.LowerIDValue;

            tResult = readItem.SqlSelectByLowerId("FloorItem", myConnection);

            if (tResult != IdLibrary.TriResult.SUCCESS)
            {
                logger.Error("InsertSampleFloor: Failed SQL select");
                myConnection.Close();
                return;
            }

            readItem.Logit(LogLevel.Info, "", "SampleFloor read in from database:", IdLibrary.LogDocDepth.MAX);

            // Clean out tables, if requested
            if (aDeleteAtEnd == true)
            {
                myTrans = myConnection.BeginTransaction();

                // Clean our records out of table
                tResult = readItem.DeleteRecordByLowerId("FloorItem", myConnection, myTrans);

                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Error("InsertSampleFloor: Failed delete flooritem and subordinate records");
                    myTrans.Rollback();
                    myConnection.Close();
                    return;
                }

                // Delete the tables
                myTrans.Commit();
                myTrans = myConnection.BeginTransaction();

                tResult = FloorItem.DropTable("FloorItem", myConnection, myTrans);

                if (tResult != IdLibrary.TriResult.SUCCESS)
                {
                    logger.Error("InsertSampleFloor: Failed drop flooritem and subordinate tables");
                    myTrans.Rollback();
                    myConnection.Close();
                    return;
                }

                // Forcing error!
                //logger.Error("InsertSampleFloor: Forcing rollback of dropping tables - empty tables should be there");
                //myTrans.Rollback();
                //myConnection.Close();
                //return;

                myTrans.Commit();
            }

            // Done with database - success
            myConnection.Close();

            // Write out result to output file for comparison
            result = readItem.WriteFile("dbsampleFloor.xml");

            if (result == false)
            {
                logger.Error("InsertSampleFloor: Failed to write out dbsampleFloor");
                return;
            }

            logger.Info("InsertSampleFloor: returning success");
        }
    }
}
